#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import numpy as np
import pandas as pd

import general as gf
from db import raw_data as rawd
from setup import ReadData as readd


class DataReader:

    __slots__ = ('_input_file', '_path', '_data_filenames', '_number_data', '_data')

    def __init__(self):
        self._path = readd.DATA_PATH
        self._data_filenames = readd.FILENAMES
        self._number_data = readd.AMOUNT_DATA
        self._data = pd.DataFrame()

    def main(self):
        """Main function to handle the input reading from the file."""
        if self._data_filenames:
            for ind, file in enumerate(self._data_filenames):
                self._read_data(file)
                if self._number_data:
                    self._check_number_of_data(ind)
                self._get_type()

    def _read_data(self, filename: str):
        """Checks which filetype is given. Distinguishes between json, excel and csv (respectively text files)"""
        filename = filename.strip()
        file = self._path + "/" + filename
        end = file[-4:]
        if end == 'xcel':
            self._data = pd.read_excel(file)
        elif end[-3:] == 'csv':
            self._data = pd.read_csv(file, sep=None, engine='python', warn_bad_lines=True)
        else:
            print("Filetype for input data not supported. Use .xcel or .csv file.")
            exit(3)

    def _check_number_of_data(self, ind: int):
        """Checks whether all lines of a file were read in successfully."""
        if len(self._data) != self._number_data[ind]:
            print("Not all data could be read in the following file: ", self._data_filenames[ind])
            print("Number of data in file: ", len(self._data))
            print("Expected amount of data:", self._number_data[ind])
            exit(11)

    def _get_type(self):
        """Decides if it is a reference list or a data list and initiate the addition to the SQL database."""
        col = list(self._data.columns)
        if "Title" in col:
            AddRef(self._data).main()
        else:
            AddData(self._data).main()


class CompletenessChecker:
    """Checks if the essential data are available. If not, the concerned data are dropped."""

    __slots__ = ('_data', '_deleted_data', '_essential_parameters')

    def __init__(self, data:pd.DataFrame):
        self._data = data
        self._essential_parameters = ['DiffusingElement', 'Alloy', 'Composition', 'Phase', 'D', 'Temp','Original']
        self._deleted_data = {}

    def main(self) -> pd.DataFrame:
        self._check()
        self._print_results()
        return self._data

    def _check(self):
        for header in self._essential_parameters :
            length_before = len(self._data)
            if header == 'D':
                self._data.drop(
                    self._data[self._data['D'].isnull() & (self._data['D0'].isnull() | self._data['Q'].isnull())].index,
                    inplace=True)
            elif header == 'Temp':
                self._data.drop(self._data[self._data['Tmin'].isnull() | self._data['Tmax'].isnull()].index,
                                inplace=True)
            else:
                self._data.drop(self._data[self._data[header].isnull()].index, inplace=True)
            self._data.reset_index(inplace=True, drop=True)
            self._deleted_data[header] = length_before - len(self._data)

    def _print_results(self):
        for par in self._essential_parameters:
            if self._deleted_data[par] > 0:
                print(self._deleted_data[par], "data were deleted because of missing ", par, ".")


class PlausibilityChecker:
    """Checks if the entered values are plausible."""
    __slots__ = '_data'

    def __init__(self, data: pd.DataFrame):
        self._data = data

    def main(self) -> pd.DataFrame:
        self._check_comp()
        self._check_reference_existence()
        self._check_number_dp()
        return self._data

    def _check_comp(self):
        path = gf.General.PATH
        filename = "/Errors_TracerData.txt"
        with open(path + filename, gf.General.get_file_modifier(path, filename)) as file:
            ind = []
            for i in range(len(self._data)):
                row = self._data.iloc[i]
                comp = row.Composition[1:-1].split(',')
                comp = [float(i) for i in comp]
                s = sum(comp)
                if abs(s-100) < 10:
                    s = s/100
                    comp = [(i/100.0) for i in comp]
                if abs(s - 1) > 1E-2:
                    for element in row:
                        file.write(str(element)+ '  ')
                    file.write("\n \n \n \n")
                    ind.append(i)
                self._data.at[i, 'Composition'] = comp
            if ind:
                self._data.drop(ind, inplace=True)
                self._data.reset_index(drop=True, inplace=True)

    def _check_number_dp(self):
        """If number of datapoints is not given, this function will put 1 for single point data and 2 if
        Arrhenius Parameter are given"""
        for i in range(len(self._data)):
            dp = self._data.loc[i, 'DP']
            if type(dp) == float and np.isnan(dp):
                d0 = self._data.loc[i, 'D0']
                if np.isnan(d0):
                    self._data.at[i, 'DP'] = 1
                else:
                    self._data.at[i, 'DP'] = 2

    def _check_reference_existence(self):
        """Checks if the reference exists in the Reference-DB for the data that should be added."""
        refs = list(set(self._data['Original'].tolist() + self._data['Taken from'].tolist()))
        id_table = self._get_refs_from_db()
        refs_to_delete = []
        ind_to_delete = []
        for i in range(len(refs)):
            if not refs[i] in id_table:
                ind_to_delete.append(i)
                refs_to_delete.append(refs[i])
        if ind_to_delete:
            self._missing_ref_warning(refs_to_delete)
            self._data.drop(ind_to_delete, inplace=True)
            self._data.reset_index(drop=True, inplace=True)

    def _missing_ref_warning(self, black_list:list):
        """Writes a file with the missing references."""
        path = gf.General.PATH
        filename = "/MissingReferences.txt"
        with open(path + filename, gf.General.get_file_modifier(path, filename)) as file:
            for ref in black_list:
                file.write(ref + "\n")

    def _get_refs_from_db(self) -> list:
        """Collects the references from the Reference-DB."""
        res = rawd.GetRef().main()
        return res.ID.tolist()


class DuplicateChecker:
    """Checks if the data contain duplicates and if the data exist already in the database."""
    __slots__ = ('_gen_data', '_comp_data')

    def __init__(self, gen_data: pd.DataFrame, comp_data: list):
        self._gen_data = gen_data
        self._comp_data = comp_data

    def main(self) -> (pd.DataFrame, list):
        rawd.RawDataAdder(self._comp_data, self._gen_data).main()
        data = rawd.GetData().main()
        data.drop_duplicates(inplace=True)
        data.reset_index(inplace=True)
        id = range(len(data))
        data.rename(columns={'DP': "Data Points"}, inplace=True)
        self._gen_data = data[['DiffusingElement', 'Tmin', 'Tmax', 'Phase', 'D', 'D0', 'Q', 'Data Points',  'Taken from',
                         'Original', 'Taken from Plot', 'Details']].copy()
        self._gen_data['ID'] = id
        self._comp_data=data.copy()
        self._comp_data.drop(['index', 'DiffusingElement', 'Tmin', 'Tmax', 'Phase', 'D', 'D0', 'Q', 'Data Points',  'Taken from',
                         'Original', 'Taken from Plot', 'Details'],axis=1, inplace=True)
        self._comp_data['ID'] = id
        return self._gen_data, self._comp_data


class PrepDataForStorage:
    """Prepares data for storage in SQL Database."""
    __slots__ = ('_data', '_gen_data', '_comp_data')

    def __init__(self, data: pd.DataFrame):
        self._data = data
        self._comp_data = []

    def main(self) -> (pd.DataFrame, list):
        self._get_comp()
        self._get_gen_data()
        return self._gen_data, self._comp_data

    def _get_comp(self):
        """Adds the composition to the Database."""
        for i in range(len(self._data)):
            row = self._data.iloc[i]
            al = self._data.loc[i, 'Alloy']
            alloy = al
            if isinstance(al, str):
                alloy = al[1:-1].split(',')
            for j in range(len(alloy)):
                if alloy[j].upper() == "AS":
                    alloy[j] = "ARS"
                if alloy[j].upper() == "IN":
                    alloy[j] = "IND"
            comp = self._data.loc[i, 'Composition'].copy()
            if len(alloy) != len(comp):
                print("Composition array and alloy array do not match in the following data set: ", row)
                exit(1)
            comp = dict(zip(alloy, comp))
            self._comp_data.append(comp)

    def _get_gen_data(self):
        """Adds the general data to the database."""
        self._data.rename(columns= {'DP': "Data Points"}, inplace=True)
        df = self._data[['DiffusingElement', 'Tmin', 'Tmax', 'Phase', 'D', 'D0', 'Q', 'Data Points',  'Taken from',
                         'Original', 'Taken from Plot', 'Details']].copy()
        self._gen_data = self._to_upper(df)

    def _to_upper(self, df:pd.DataFrame)->pd.DataFrame:
        """Changes the given strings in the given columns to upper case strings."""
        header = ['DiffusingElement', 'Phase', 'Taken from', 'Original', 'Taken from Plot', 'Details']
        for h in header:
            s = df[h]
            df[h] = s.str.upper()
        return df


class DeleteAssRes:
    """Marks assessment results that have to be reassessed due to the newly added data."""
    __slots__ = ('_data', '_elements', '_phase', '_databases')

    def __init__(self, data: pd.DataFrame):
        self._data = data
        self._elements = self._data['DiffusingElement'].tolist()
        self._phase = self._data['Phase'].tolist()
        self._databases = ["SQL_Databases/AssessmentResults/Endmember.db",
                           "SQL_Databases/AssessmentResults/BinaryInteractions.db",
                           "SQL_Databases/AssessmentResults/TernaryInteractions.db"]

    def main(self):
        """Tag the assessment results that have to be reassessed."""
        for db in self._databases:
            with gf.General.create_connection(gf.General.PATH, db) as conn:
                self._delete_ass_res(conn)

    def _delete_ass_res(self, conn):
        cur = conn.cursor()
        for i in range(len(self._elements)):
            sql = """DELETE FROM DataInformation WHERE "DiffusingElement"="{el}" AND "Phase"="{ph}" ; """.format(
                el=self._elements[i].strip().upper(), ph=self._phase[i].strip().upper())
            cur.execute(sql)
            conn.commit()


class AddData:

    __slots__ = '_data'

    def __init__(self, data: pd.DataFrame):
        self._data = data
        self._data.replace(to_replace=["Null", "NULL", "null", None], value=np.nan, inplace=True, regex=True)

    def main(self):
        self._rename()
        self._data = CompletenessChecker(self._data).main()
        self._data = PlausibilityChecker(self._data).main()
        DeleteAssRes(self._data).main()
        gen_data, comp_data = PrepDataForStorage(self._data).main()
        gen_data_final, comp_data_final = DuplicateChecker(gen_data, comp_data).main()
        rawd.RawDataAdder.readd_data(comp_data_final, gen_data_final)

    def _rename(self):
        self._data.rename(columns= {"Tmin [K]": "Tmin", "Tmax [K]": "Tmax", 'D[m²/s]': "D",'D0[m²/s]': "D0",
                                    'Q[J/mol]': "Q", 'Number data Points': 'DP'}, inplace=True)
        self._data[['Tmin', 'Tmax', 'D', 'D0', 'Q', 'DP']] = self._data[['Tmin', 'Tmax', 'D', 'D0', 'Q', 'DP']].astype(float)


class AddRef:
    """Handles addition of new references to the SQL database.

    Data are checked for completeness, if the correct ID was given by the user and for duplicates.
    """

    __slots__ = ('_data', '_deleted_data')

    def __init__(self, data:pd.DataFrame):
        self._data = data
        self._data.replace(to_replace=["Null", "NULL", "null", None], value=np.nan, inplace=True, regex=True)
        self._deleted_data = {}

    def main(self):
        self._check_completeness()
        self._check_id()
        self._merge_old_new_data()
        self._drop_duplicates()
        self._print_results()
        rawd.RefAdder(self._data).main()

    def _drop_duplicates(self):
        length_before = len(self._data)
        self._data.drop_duplicates(inplace=True)
        length_after = len(self._data)
        self._data.drop_duplicates(['ID'], inplace=True)
        self._data.reset_index(inplace=True, drop=True)
        self._deleted_data['duplicates'] = length_before - length_after

    def _check_completeness(self):
        length_before_ID = len(self._data)
        self._data.drop(self._data[self._data['ID'].isnull()].index, inplace=True)
        length_before_year = len(self._data)
        self._deleted_data['ID_null'] = length_before_ID - length_before_year
        self._data.drop(self._data[self._data['Year'].isnull()].index, inplace=True)
        length_before_name = len(self._data)
        self._deleted_data['Year_null'] = length_before_year - length_before_name
        self._data.drop(self._data[self._data['First author (last name)'].isnull()].index, inplace=True)
        length_after_name = len(self._data)
        self._deleted_data['author_null'] = length_before_name - length_after_name

    def _check_id(self):
        drop_ind = []
        self._data.reset_index(inplace=True, drop=True)
        for ind in range(len(self._data)):
            ID = self._data.loc[ind, 'ID'].upper().strip()
            year = self._data.loc[ind, 'Year']
            name = self._data.loc[ind, 'First author (last name)'].upper().strip()
            test_ID = name +str(year)
            if ID != test_ID and ID[:-1] != test_ID:
                drop_ind.append(ind)
        print("The following references did not match ID and author name + year:")
        for line in drop_ind:
            print(self._data.iloc[line])
        self._data.drop(drop_ind, inplace=True)
        self._data.reset_index(inplace=True, drop=True)
        self._deleted_data['wrong_ID'] = len(drop_ind)

    def _merge_old_new_data(self):
        old_data = rawd.GetRef().main()
        if not old_data.empty:
            self._data.append(old_data, sort=False)

    def _print_results(self):
        print(self._deleted_data['duplicates'], " references were deleted because they were duplicates.")
        print(self._deleted_data['ID_null'], " references were deleted because the ID was Null.")
        print(self._deleted_data['Year_null'], " references were deleted because the Year was Null.")
        print(self._deleted_data['author_null'], " references were deleted because the first author's name was Null.")
        print(self._deleted_data['wrong_ID'], " references were deleted because the ID was wrong.")






