#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import collections

import pandas as pd

import general as gf
from assessment import errors as em
from assessment import fitting as fit
from assessment.assessment_parameters import AssPar as ap
from db import ass_res as ar
from output import binary_plot as bp


class InitBinary:
    """Initializes the binary assessments."""
    phase = None

    __slots__ = ('_alloy', 'ass_obj_list')

    def __init__(self, phase: str):
        self._alloy = ap.alloy
        self.ass_obj_list = collections.OrderedDict()
        InitBinary.phase = phase

    def main(self):
        self._initialize()
        for key in self.ass_obj_list:
           self.ass_obj_list[key].ass_main()

    def _initialize(self):
        """Initializes all binary alloying systems that have to be assessed."""
        count = 0
        for el in self._alloy:
            for inter in self._alloy:
                if inter != el:
                    self.ass_obj_list[count] = BinaryInteraction(el, el, inter)
                    count = count + 1
        for el in self._alloy:
            for inter1 in self._alloy:
                if el != inter1:
                    for inter2 in self._alloy:
                        if inter2 != el and inter2 > inter1:
                            self.ass_obj_list[count] = BinaryInteraction(el, inter1, inter2)
                            count = count + 1


class BinaryInteraction:
    """Handles the binary assessment."""

    __slots__ = ('el', '_inter1', '_inter2', 'assessed', '_ass_par_un', '_data_list', '_res', '_data')

    def __init__(self, el: str, inter1: str, inter2: str):
        self.el = gf.General.clean_input(el)
        self._inter1 = gf.General.clean_input(inter1)
        self._inter2 = gf.General.clean_input(inter2)
        self.assessed = False
        self._res = pd.DataFrame(columns=ap.BIN_RES_HEADER)
        self._ass_par_un = {}
        self._data_list = pd.DataFrame()
        self._data = pd.DataFrame()

    def ass_main(self):
        """Decides whether an assessment is needed or assessed data can be loaded."""
        self._get_ass_data()
        if not self.assessed:
            self._assess()

    def _get_ass_data(self):
        """Returns the assessed data from the database for the given element in the given alloy."""
        res = ar.GetAssRes().get_bin()
        if not res.empty:
            res.drop(res[(res['DiffusingElement'] != self.el) | (res['IntEl1'] != self._inter1) |
                         (res['IntEl2'] != self._inter2) | (res['Phase'] != InitBinary.phase)].index, inplace=True)
            res.reset_index(inplace=True, drop=True)
        if not res.empty:
            self.assessed = True
            res.rename(columns={'DiffusingElement': 'DiffEl'}, inplace=True)
            self._res = res
            self._store_results()

    def _assess(self):
        """Handles the assessment."""
        self._select_data()
        self._get_lower_data()
        if not self._data.empty:
            self._data_list = gf.PandasRelated.create_data_list(self._data)
            temp_res = self._fit()
            self._select_best_res(temp_res)
            if self.el == self._inter1:
                self._plot()
            self.assessed = True
            self._store_results()
        else:
            self._save_default()

    def _get_al(self) -> list:
        al = [self.el, self._inter1, self._inter2]
        if self.el == self._inter1:
            al = [self.el, self._inter2]
        al.sort()
        return al

    def _select_data(self):
        """Selects available data for binary assessment."""
        dat = ap.temp_data.copy()
        dat = gf.PandasRelated.select_data_by_diff_elem(self.el, dat)
        dat = gf.PandasRelated.select_data_by_alloy(self._get_al(), dat)
        if self.el != self._inter1:
            dat = gf.PandasRelated.select_imp_data(self.el, dat)
        else:
            dat = gf.PandasRelated.select_self_diff_data(self.el, dat)
        if not dat.empty:
            dat.reset_index(inplace=True, drop=True)
        self._data = dat

    def _get_lower_data(self):
        """Collects unary data."""
        self._ass_par_un = gf.AssResToDic.un_to_dict(ap.get_unary_results(), InitBinary.phase)

    def _fit(self) -> pd.DataFrame:
        """Handles fitting of binary alloys."""
        max_order = ap.BINARY_ORDER
        res = dict()  # keys are the order and without an excess term the order is taken as -1. As parameters only excess parameters are saved
        result = gf.LnDfromAss(self.el, self._data_list, InitBinary.phase, self._ass_par_un).main()
        basic_num_par = 2
        err = em.ErrorManager(basic_num_par, self._data_list['lnD'].tolist(), result.tolist())
        err.get_errors()
        res[-1] = {'parameter': [], 'BIC': err.results['BIC'], 'AIC': err.results['AIC']}

        for i in range(max_order):
            if i+basic_num_par+3 > len(self._data_list):
                print("Too many parameters")
                break
            excess_par = [0.0] * (max_order+1)
            fitres = fit.FitBin(InitBinary.phase, self._data_list, result, order=i, diff_el=self.el,
                                int_el1=self._inter1, int_el2=self._inter2).main()

            if fitres is not None:
                ass_par_temp = dict(self._ass_par_un, **fitres)
                pred_bin = gf.LnDfromAss(self.el, self._data_list, InitBinary.phase, ass_par_temp).main()
                num_par = basic_num_par + 1 + i
                err = em.ErrorManager(num_par, self._data_list['lnD'].tolist(), pred_bin)
                err.get_errors()
                count = 0
                for key in fitres:
                    excess_par[count] = fitres[key]
                    count = count + 1
                res[i] = {'parameter': excess_par, 'BIC': err.results['BIC'], 'AIC': err.results['AIC']}
        temp_res = self._add_local_res(res)
        return temp_res

    def _add_local_res(self, res: dict) -> pd.DataFrame:
        """Store the parameters locally."""
        temp_res = pd.DataFrame(columns=ap.BIN_RES_HEADER)
        for order in res:
            if order == -1:
                temp_res = temp_res.append({
                'HighestOrder': order, 'DiffEl': self.el, 'IntEl1': self._inter1, 'IntEl2': self._inter2,
                'Order': order, 'MQlowT': 0.0, 'BP': None, 'MQhighT': 0.0, 'BIC': res[order]['BIC'],
                'AIC': res[order]['AIC'], 'Mode': 'UNARY', 'DP': len(self._data_list)}, ignore_index=True)
            else:
                par = res[order]['parameter'].copy()
                for i in range(order + 1):
                    temp_res = temp_res.append({
                    'HighestOrder': order, 'DiffEl': self.el, 'IntEl1': self._inter1, 'IntEl2': self._inter2,
                    'Order': i, 'MQlowT': par[i], 'BP': None, 'MQhighT': par[i], 'BIC': res[order]['BIC'],
                    'AIC': res[order]['AIC'], 'Mode': 'NORMAL', 'DP': len(self._data_list)}, ignore_index=True)
        return temp_res

    def _select_best_res(self, temp_res: pd.DataFrame):
        """Selects the best result."""
        error = ap.SELECTION_CRITERION
        if len(temp_res) > 1:
            min = temp_res[error].min()
            if isinstance(min, list):
                min = min[0]
            ho = temp_res.loc[temp_res[error] == min, 'HighestOrder']
            ho = ho.iloc[0]
            res = temp_res.loc[temp_res['HighestOrder'] == ho, :]
            res.reset_index(drop=True, inplace=True)
            self._res = res
        else:
            self._res = temp_res

    def _save_default(self):
        """Sets default values if an assessment is not possible."""
        def_res = pd.DataFrame(columns=ap.BIN_RES_HEADER)
        def_res = def_res.append({'HighestOrder': -1, 'Parameter': 'MQ', 'DiffEl': self.el, 'IntEl1': self._inter1, 'IntEl2':  self._inter2,
            'Order': -1, 'MQlowT': 0, 'MQhighT': 0, 'BP': None, 'AIC': 0, 'BIC': 0, 'R2_adj': 0, 'Mode': 'No data','DP': 0}, ignore_index=True)
        ap.add_binary_result(def_res)

    def _plot(self):
        """Plots the comparison for the excess parameters."""
        bp.BinaryPlot(self.el, self._inter1, self._inter2, InitBinary.phase, self._data, self._ass_par_un,
                      gf.AssResToDic.bin_to_dict(self._res, InitBinary.phase)).main()

    def _store_results(self):
        """Adds data of the given system to the local storage."""
        if not self._res.empty:
            ap.add_binary_result(self._res)

