#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import collections

import pandas as pd

import general as gf
from assessment import fitting as fit
from assessment import od as od
from assessment.assessment_parameters import AssPar as ap
from db import ass_res as ar
from output import binary_plot as bp
from output import endmember_plot as un


class InitUnary:
    """Initializes the assessment of the endmember data."""
    data = pd.DataFrame()
    phase = None
    not_ass_list = []

    __slots__ = ('_alloy', 'ass_obj_list')

    def __init__(self, phase: str):
        self._alloy = ap.alloy
        self.ass_obj_list = collections.OrderedDict()
        InitUnary.data = ap.temp_data
        InitUnary.phase = phase

    def main(self):
        """Handles the assessment of all unaries"""
        self._initialize()
        for key in self.ass_obj_list:
            self.ass_obj_list[key].ass_main()
        self._assess_from_higher_order()

    def _initialize(self):
        """Calls all unaries for that alloy."""
        for i in range(len(self._alloy)):
            for j in range(len(self._alloy)):
                diff_el = self._alloy[i]
                mat = self._alloy[j]
                k = diff_el + "_in_" + mat
                self.ass_obj_list[k] = Unary(diff_el, mat)

    def _assess_from_higher_order(self):
        """Assesses the missing data from higher order systems."""
        InitUnary.not_ass_list.sort()
        not_ass_list_temp = InitUnary.not_ass_list
        InitUnary.not_ass_list = []
        for i in range(len(not_ass_list_temp)):
            not_ass_list_temp[i].main()
        for i in range(len(not_ass_list_temp)):
            if not not_ass_list_temp[i].assessed:
                not_ass_list_temp[i].set_default()


class Unary:
    """Assessment for Unary objects."""

    __slots__ = ('el', 'mat', '_ass_detail', 'assessed', '_res', '_data')

    def __init__(self, el, al):
        self.el = gf.General.clean_input(el)
        self.mat = gf.General.clean_input(al)
        self._ass_detail = "Normal"
        self.assessed = False
        self._res = pd.DataFrame(columns=ap.UN_RES_HEADER)
        self._data = pd.DataFrame()

    def __lt__(self, other):
        if self.el == self.mat and other.el != other.mat:
            return True
        elif self.el != self.mat and other.el == other.mat:
            return False
        elif self.el != self.mat and other.el != other.mat:
            return (self.el, self.mat) < (other.el, other.mat)
        elif self.el == self.mat and other.el == other.mat:
            return self.el < other.el
        else:
            print("Comparison case does not exist")
            exit(6)

    def ass_main(self):
        """Decides whether an assessment is needed or assessed data can be loaded."""
        self._check_existing_result()
        if not self.assessed:
            self._assess()

    def _check_existing_result(self):
        """Returns the assessed data from the database for the given element in the given alloy."""
        res = ar.GetAssRes().get_un()
        res.drop(res[(res['DiffusingElement'] != self.el) | (res['Matrix'] != self.mat) | (res['Phase'] != InitUnary.phase)].index, inplace=True)
        res.reset_index(inplace=True, drop=True)
        if not res.empty:
            res.rename(columns={'DiffusingElement': 'DiffEl'}, inplace=True)
            self._res = res
            self.assessed = True
            self._store_results()

    def _assess(self):
        """Handles the assessment"""
        self._select_data()
        if not self._data.empty:
            self._outlier_det()
            self._fit()
            self._plot_fit()
            self.assessed = True
            self._store_results()
        else:
            mode = "SELF"
            if self.el != self.mat:
                mode = "IMP"
            InitUnary.not_ass_list.append(UnFromHigher(mode, self.el, self.mat))

    def _select_data(self):
        """Select the raw data for the Assessment"""
        dat = ap.temp_data.copy()
        self._data = gf.PandasRelated.select_data_by_diff_elem(self.el, dat)
        self._data.drop(self._data[self._data[self.mat] != 1.0].index, inplace=True)
        self._data.reset_index(drop=True, inplace=True)

    def _outlier_det(self):
        """Checks whether outlier detection for unaries is needed and then if necessary performs outlier detection."""
        if ap.od:
            if od.ODA.test_outlier_detection(self._data):
                self._data = od.ODA.od_main(self._data)

    def _fit(self):
        """Checks if one or two Arrhenius plots fit the data best"""
        data_list = gf.PandasRelated.create_data_list(self._data)
        if "weight" in list(self._data.columns.values):
            res = fit.FitUn(self.el, self.mat, InitUnary.phase, data_list, "UNARY", weights=self._data['weight']).main()
        else:
            res = fit.FitUn(self.el, self.mat, InitUnary.phase, data_list, "UNARY").main()
        key_mq = self.el + "_" +"MQ"
        key_mf = self.el + "_" +"MF"
        mq = res[key_mq]
        mf = res[key_mf]
        self._res = self._res.append({
            'MFlowT': mf, 'MQlowT': mq, 'BP': None, 'MFhighT': mf, 'MQhighT': mq, 'DiffEl': self.el, 'Matrix': self.mat,
            'Mode': self._ass_detail, 'DP': len(self._data)}, ignore_index=True)

    def _plot_fit(self):
        """Plots the fit against the data. Also plots the new assessed data against old assessed data if possible."""
        un.UnPl(self.el, self.mat, InitUnary.phase, self._data, self._res).main()

    def _store_results(self):
        """Adds data of the given system to the local storage"""
        ap.add_unary_result(self._res)


class UnFromHigher:
    """Deals with the assessment of unaries from higher order systems"""

    __slots__ = ('_mode', 'el', 'mat', '_res', '_data', '_un_dat', 'assessed', '_ass_detail')

    def __init__(self, mode, el, mat):
        self._mode = mode
        self.el = gf.General.clean_input(el)
        self.mat = gf.General.clean_input(mat)
        self._res = pd.DataFrame(columns=ap.UN_RES_HEADER)
        self._data = None
        self._un_dat = pd.DataFrame
        self.assessed = False
        self._ass_detail = None

    def __str__(self):
        return "UnFromHigher: " + self.el + " in " + self.mat

    def __lt__(self, other):
        if self.el == self.mat and other.el != other.mat:
            return True
        elif self.el != self.mat and other.el == other.mat:
            return False
        elif self.el != self.mat and other.el != other.mat:
            return (self.el, self.mat) < (other.el, other.mat)
        elif self.el == self.mat and other.el == other.mat:
            return self.el < other.el
        else:
            print("Comparison case does not exist")
            exit(6)

    def main(self):
        self._get_raw_data()
        if not self._data.empty:
            self._get_un_data()
        if not self._data.empty and not self._un_dat.empty:
            self._minimize()
        if not self._res.empty:
            self._plot()
        else:
            InitUnary.not_ass_list.append(self)

    def _get_raw_data(self):
        """Collects all data that are used for this assessment"""
        dat = ap.temp_data.copy()
        if not dat.empty:
            if self._mode == "SELF":
                dat = gf.PandasRelated.select_data_by_diff_elem(self.el, dat)
                self._data = gf.PandasRelated.select_only_bin_data(dat)
            else:
                dat = gf.PandasRelated.select_data_by_diff_elem(self.el, dat)
                al = [self.el]
                if self.el != self.mat:
                    al = [self.el, self.mat]
                    al.sort()
                self._data = gf.PandasRelated.select_data_by_alloy(al, dat)

    def _get_un_data(self):
        """Collects all already assessed unary data"""
        self._un_dat = ap.get_unary_results()
        if self._mode == "IMP":
            self._un_dat.drop(self._un_dat[(self._un_dat['DiffEl'] != self.el) | (self._un_dat.Matrix != self.el)].index, inplace=True)
            self._un_dat.reset_index(drop=True, inplace=True)
        elif self._mode == "SELF":
            self._un_dat.drop(self._un_dat[self._un_dat['DiffEl'] != self.el].index, inplace=True)
            self._un_dat.reset_index(drop=True, inplace=True)

    def _minimize(self):
        """Determines the assessment parameters."""
        data_list = gf.PandasRelated.create_data_list(self._data)
        data_list.drop_duplicates(inplace=True)
        data_list.reset_index(inplace=True, drop=True)
        mode = "UN_FROM_HIGHER"
        fit_par = fit.FitUn(self.el, self.mat,InitUnary.phase, data_list, mode, self._un_dat).main()

        if fit_par is not None:
            self.assessed = True
            self._ass_detail = "HIGHER_ORDER"
            self._res = self._res.append({
                'MFlowT': fit_par['mf'], 'MQlowT': fit_par['mq'], 'BP': None, 'MFhighT': fit_par['mf'],
                'MQhighT': fit_par['mq'],  'DiffEl': self.el,  'Matrix': self.mat, 'Mode': self._ass_detail,
                'DP': len(data_list)}, ignore_index=True)
            self._store_results()

    def _plot(self):
        """Plots the assessed data from the impurity fit for the unary data"""
        if self._mode == "IMP":
            bp.BinaryPlot(self.el, self.el, self.mat, InitUnary.phase, self._data, gf.AssResToDic.un_to_dict(self._res,
                        InitUnary.phase), None).main()

    def _store_results(self):
        """Adds data of the given system to the local storage"""
        ap.add_unary_result(self._res)

    def set_default(self):
        """Sets for M0 and Q default values since data are missing."""
        if self._res.empty and not self.assessed:
            self._ass_detail = "DEFAULT"
            if self.el == self.mat:
                self._set_default_self()
            else:
                self._adapt_matrix_res()
            self.assessed = True
            self._store_results()

    def _adapt_matrix_res(self):
        """If there are no unary/binary data for diffusion of A in B, this will be set equal to diffusion of B in B."""
        un_res = ap.get_unary_results()
        try:
            un_res.loc[(un_res['DiffEl'] == self.mat) & (un_res['Matrix'] == self.mat), 'MFlowT'].item()
        except ValueError:
            pass
        else:
            MFlowT = un_res.loc[(un_res['DiffEl'] == self.mat) & (un_res['Matrix'] == self.mat), 'MFlowT'].item()
            MQlowT = un_res.loc[(un_res['DiffEl'] == self.mat) & (un_res['Matrix'] == self.mat), 'MQlowT'].item()
            MFhighT = un_res.loc[(un_res['DiffEl'] == self.mat) & (un_res['Matrix'] == self.mat), 'MFhighT'].item()
            MQhighT = un_res.loc[(un_res['DiffEl'] == self.mat) & (un_res['Matrix'] == self.mat), 'MQhighT'].item()
            breakpoint = un_res.loc[(un_res['DiffEl'] == self.mat) & (un_res['Matrix'] == self.mat), 'BP'].item()
            self.assessed = True
            self._ass_detail = "ADOPTED"
            self._res = self._res.append({
                'MFlowT': MFlowT, 'MQlowT': MQlowT, 'BP': breakpoint, 'MFhighT': MFhighT, 'MQhighT': MQhighT,
                'DiffEl': self.el, 'Matrix': self.mat, 'Mode':  self._ass_detail, 'DP': len(self._data)},
                ignore_index=True)

    def _set_default_self(self):
        """Sets a default value for missing self-diffusion data."""
        MF_def = 1E-5
        mp = gf.PhysicalProperties.get_mp(self.el)
        if mp is not None:
            MQ_def = 8.314*17*mp
        else:
            MQ_def = 300000
        dp = 0
        self._res = self._res.append({
            'MFlowT': MF_def, 'MQlowT': MQ_def, 'BP': None, 'MFhighT': MF_def, 'MQhighT': MQ_def, 'DiffEl': self.el,
            'Matrix': self.mat, 'Mode': self._ass_detail, 'DP': dp}, ignore_index=True)


