#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import lmfit
import numpy as np
import pandas as pd

import general as gf

from lmfit import Parameters, fit_report, minimize
class Fit:
    """Superclass for fitting. Contains functions and initializations used fin all fittin classes."""
    R = 8.314

    __slots__ = '_data_list'

    def __init__(self, data_list: pd.DataFrame):
        self._data_list = data_list

    def _get_el_list(self) -> list:
        """Create a list of elements."""
        l = list(self._data_list.columns)
        l.remove('lnD')
        l.remove('Temp')
        if 'DiffEl' in l:
            l.remove('DiffEl')
        return l


class FitUn(Fit):
    """Handles fitting of endmember data."""

    __slots__ = ('_el', '_mat', '_ass_par_un', '_mode', '_el_l', '_weights', '_phase')

    def __init__(self, el, mat, phase, data_list, mode, ass_par_un=None, weights=None):
        super(FitUn, self).__init__(data_list)
        self._el = el
        self._mat = mat
        self._phase = phase
        self._ass_par_un = ass_par_un #dict: key is element and contains dict with parameters
        self._mode = mode
        self._el_l = None
        self._weights = [1.0]*len(self._data_list)
        if weights is not None:
            self._weights = [1/w for w in weights]

    def main(self):
        if self._mode == "UNARY" and not self._data_list.empty:
            return self._unary_normal()
        if self._mode == "UN_FROM_HIGHER" and not self._data_list.empty:
            return self._un_from_higher()
        else:
            return None

    def _un_from_higher(self):
        """Handles fitting of endmember data from binary data."""
        self._el_l = self._get_el_list()
        params = lmfit.Parameters()
        mf = 1E-5
        mq = 3E5
        params.add("mf", value=mf, min=1E-10, max=1, vary=True)
        params.add("mq", value=mq, min=1E5, max=1E7, vary=True)

        def func(params, x, lnD=None, eps=None):
            """Creates the fitting function."""
            temp = np.asarray(x['Temp']).astype(float)
            fun = gf.LnDfromAss(self._el, self._data_list, self._phase, gf.AssResToDic.un_to_dict(self._ass_par_un, self._phase)).main()
            mf = params["mf"]
            mq = params["mq"]
            comp = np.asarray(x[self._el]).astype(float)
            func = comp*((Fit.R * temp * np.log(mf) - mq) / (Fit.R * temp))
            if fun.empty:
                fun = func
            else:
                fun = fun + func
            if lnD is None:
                print("wenn ich hier bin, bin ich falsch")
                exit(31)
                return fun
            if eps is None:
                diff = fun - lnD
                return diff.values.tolist()
            r = (fun - lnD) / eps
            return r.tolist()

        x = self._data_list.copy()
        x.drop(['lnD'], axis=1, inplace=True)
        try:
            res = lmfit.minimize(func, params, args=(x, self._data_list['lnD']))
            res = res.params.valuesdict()
        except ValueError:
            res = None
        return res

    def _unary_normal(self):
        """Handles fitting of normal endmember data."""
        params = lmfit.Parameters()
        key_mf = self._el + "_MF"
        key_mq = self._el + "_MQ"
        mf = 1E-5
        mq = 3E5
        params.add(key_mf, value=mf, min=1E-10, max=1, vary=True)
        params.add(key_mq, value=mq, min=1E5, max=1E7, vary=True)

        def func(par, x, lnD=None, eps=None):
            temp = x['Temp']
            fun = pd.Series()
            mf = par[self._el + "_MF"]
            mq = par[self._el + "_MQ"]
            f = (FitUn.R * temp * np.log(mf) - mq) / (FitUn.R * temp)
            if fun.empty:
                fun = f
            else:
                fun = fun + f
            if lnD is None:
                print("wenn ich hier bin, bin ich falsch")
                exit(31)
                return fun
            if eps is None:
                diff = fun-lnD
                return diff.values.tolist()
            r = (fun-lnD)/eps
            return r.tolist()

        x = self._data_list.copy()
        x.drop(['lnD'], axis=1, inplace=True)
        res = lmfit.minimize(func, params, args=(x, self._data_list['lnD'], self._weights))
        print(fit_report(res))
        return res.params.valuesdict()


class FitBin(Fit):
    """Handles fitting of binary data."""

    __slots__ = ('_phase', '_pred', '_order', '_diff_el', '_int_el1', '_int_el2')

    def __init__(self, phase, data_list, predictions, order=0, diff_el=None, int_el1 =None, int_el2=None):
        super(FitBin, self).__init__(data_list)
        self._phase = phase
        self._pred = predictions
        self._order = order
        self._diff_el = diff_el
        self._int_el1 = int_el1
        self._int_el2 = int_el2

    def main(self) -> dict:
        params = lmfit.Parameters()
        for i in range(self._order + 1):
            key = "MQ_" + self._phase + "_" + self._diff_el + "_in_" + self._int_el1 + "_" + self._int_el2 + "_" + str(i)
            params.add(key, value=0.0, vary=True)

        def func(par, x, lnD):
            temp = x['Temp']
            fun = self._pred.copy()
            for i in range(self._order +1):
                c1 = x[self._int_el1].astype(float)
                c2 = x[self._int_el2].astype(float)
                p = par["MQ_"+ self._phase + "_" + self._diff_el + "_in_" + self._int_el1 + "_" + self._int_el2 + "_" + str(i)]
                fun += c1*c2*p*(c1-c2)**i/(Fit.R*temp)
            return fun-lnD

        x = self._data_list.copy()
        x.drop(['lnD'], axis=1, inplace=True)
        try:
            res = lmfit.minimize(func, params, args=(x, self._data_list['lnD']))
            res = res.params.valuesdict()
        except ValueError:
            res = None
        return res


class FitTern(Fit):
    """Handles fitting of ternary interactions."""

    __slots__ = ('_phase', '_pred', '_el_l', '_diff_el', '_int_el1', '_int_el2', '_int_el3', '_main_el1', '_main_el2', '_main_el3')

    def __init__(self, phase, data_list, predictions, diff_el=None, main_el1=None, main_el2=None, main_el3=None,
                 int_el1=None, int_el2=None, int_el3=None):
        super(FitTern, self).__init__(data_list)
        self._phase = phase
        self._pred = predictions
        self._el_l = None
        self._diff_el = diff_el
        self._int_el1 = int_el1
        self._int_el2 = int_el2
        self._int_el3 = int_el3
        self._main_el1 = main_el1
        self._main_el2 = main_el2
        self._main_el3 = main_el3

    def main(self):
        self._el_l = self._get_el_list()

        params = lmfit.Parameters()
        if self._main_el1 is not None:
            key = "MQ_" + self._phase + "_" + self._diff_el + "_in_" + self._int_el1 + "_" + self._int_el2 + "_" + \
                  self._int_el3 + "_main_" + self._main_el1 + "_" + str(2)
            params.add(key, value=0.0, vary=True)
        if self._main_el2 is not None:
            key = "MQ_" + self._phase + "_" + self._diff_el + "_in_" + self._int_el1 + "_" + self._int_el2 + "_" + \
                  self._int_el3 + "_main_" + self._main_el2 + "_" + str(2)
            params.add(key, value=0.0, vary=True)
        if self._main_el3 is not None:
            key = "MQ_" + self._phase + "_" + self._diff_el + "_in_" + self._int_el1 + "_" + self._int_el2 + "_" + \
                  self._int_el3 + "_main_" + self._main_el3 + "_" + str(2)
            params.add(key, value=0.0, vary=True)
        def func(par, x, lnD):
            temp = x['Temp']
            fun = self._pred.copy()
            if self._main_el1 is not None:
                c1 = x[self._int_el1].astype(float)
                c2 = x[self._int_el2].astype(float)
                c3 = x[self._int_el3].astype(float)
                c_main = x[self._main_el1].astype(float)
                mu = c_main + (1 - c1 - c2 - c3) / 3
                p = par["MQ_" + self._phase + "_" + self._diff_el + "_in_" + self._int_el1 + "_" + self._int_el2 + "_"
                        + self._int_el3 + "_main_" + self._main_el1 + "_" + str(2)]
                fun += c1 * c2 * c3 * mu * p / (Fit.R * temp)
            if self._main_el2 is not None:
                c1 = x[self._int_el1].astype(float)
                c2 = x[self._int_el2].astype(float)
                c3 = x[self._int_el3].astype(float)
                c_main = x[self._main_el2].astype(float)
                mu = c_main + (1 - c1 - c2 - c3) / 3
                p = par["MQ_" + self._phase + "_" + self._diff_el + "_in_" + self._int_el1 + "_" + self._int_el2 + "_"
                        + self._int_el3 + "_main_" + self._main_el2 + "_" + str(2)]
                fun += c1 * c2 * c3 * mu * p / (Fit.R * temp)
            if self._main_el3 is not None:
                c1 = x[self._int_el1].astype(float)
                c2 = x[self._int_el2].astype(float)
                c3 = x[self._int_el3].astype(float)
                c_main = x[self._main_el3].astype(float)
                mu = c_main + (1 - c1 - c2 - c3) / 3
                p = par["MQ_" + self._phase + "_" + self._diff_el + "_in_" + self._int_el1 + "_" + self._int_el2 + "_"
                        + self._int_el3 + "_main_" + self._main_el3 + "_" + str(2)]
                fun += c1 * c2 * c3 * mu * p / (Fit.R * temp)
            return fun-lnD

        x = self._data_list.copy()
        x.drop(['lnD'], axis=1, inplace=True)
        try:
            res = lmfit.minimize(func, params, args=(x, self._data_list['lnD']))
            res = res.params.valuesdict()
        except ValueError:
            print("ValuError")
            exit(22)
            res = None
        return res

