#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import general as gf
from assessment.assessment_parameters import AssPar as ap
from assessment.binary_interaction import InitBinary
from assessment.endmembers import InitUnary
from assessment.ternary_interaction import InitTernary
from db import ass_res as ar
from db import raw_data as rd


class Assessment:
    """Initializes the parameters for the assessment and coordinates the assessment of the different layers (first
    end-members, then binary interactions and finally ternary interactions."""

    def main(self):
        self._get_data()
        self._coord_assessment()

    def _get_data(self):
        """Collect data from database."""
        data = rd.GetData().main()
        data = gf.Conv().arr_par_into_d(data)
        data = gf.DataManipulator().prepare_data(data)
        ap.data = data

    def _coord_assessment(self):
        """Loops over all phases and coordinates the assesssment of the three layers: Unary, Binary and Ternary."""
        if not ap.data.empty:
            for ph in ap.phases:
                ap.set_results()
                self._clear_data_for_phase(ph)
                if not ap.temp_data.empty:
                    InitUnary(ph).main()
                    if len(ap.alloy) > 1:
                        InitBinary(ph).main()
                        if len(ap.alloy) > 2:
                             InitTernary(ph).main()
                    print(ap.get_unary_results())
                    ar.SaveAssRes(ph, ap.get_unary_results(), ap._bin_res, ap._tern_res).main()
                else:
                    print("No data available for assessment.")

    @staticmethod
    def _clear_data_for_phase(ph: str):
        """Drops all data that do not belong to the given phase."""
        data = ap.data.copy()
        data.drop(data[data['Phase'] != ph].index, inplace=True)
        data.reset_index(drop=True, inplace=True)
        ap.temp_data = data

