#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

'''Handles Outlier Detection '''

import math as math

import numpy as np
import pandas as pd

import general as gf
from assessment import errors as em


class ODA:
    """Outlier Detection Arrhenius: conducts outlier detection for an Arrhenius fit and calculates the weights for
    the data"""
    _test_data = pd.DataFrame()
    _train_data = pd.DataFrame()
    _rse = []
    _obs = []
    _MF = []
    _MQ = []

    @classmethod
    def _get_error(cls):
        """Calculates error for OD."""
        num_par = 2
        data = ODA._test_data.lnD.values
        pred_dat = ODA._get_predicted_val()
        err = em.ErrorManager(num_par, data, pred_dat)
        err.get_RSE() #RSS
        return err.results['RSE']

    @classmethod
    def _get_predicted_val(cls):
        """Calculates the predicted values fitting to the test data."""
        R = 8.314
        dat = np.log(ODA._MF[-1])-ODA._MQ[-1]/(R*ODA._test_data.Temp)
        return dat.values

    @classmethod
    def _get_diffusion_parameters(cls):
        """Fits the train data and calculates M0 and Q from it"""
        # least squares polynomial fit returns polynomial coefficients, highest order first
        result_fit = np.polyfit(ODA._train_data['invT'].tolist(), ODA._train_data['lnD'].tolist(), deg=1, full=True, cov=True)
        ODA._MF.append(math.exp(result_fit[0][1]))
        ODA._MQ.append(-(result_fit[0][0] * 8.314))

    @classmethod
    def _outlier_detection(cls, data, refs):
        """Handles outlier detection."""
        rse_av = False
        for i in range(len(refs)):
            ODA._test_data = pd.DataFrame()
            ODA._train_data = pd.DataFrame()
            ODA._test_data = data.loc[data['Original'] == refs.iloc[i, 0]].copy()
            ODA._train_data = data.loc[data['Original'] != refs.iloc[i, 0]].copy()
            ODA._test_data.reset_index(drop=True, inplace=True)
            ODA._train_data.reset_index(drop=True, inplace=True)
            if len(ODA._train_data) >= 3: # make sure that rse can be calculated
                ODA._get_diffusion_parameters()
                ODA._rse.append(ODA._get_error())
                ODA._obs.append(len(ODA._test_data))
                rse_av = True
            else:
                ODA._rse.append(np.nan)
                ODA._obs.append(np.nan)
        return rse_av

    @classmethod
    def _get_weights(cls, rse_av, refs):
        """Calculates the weights for each sample set."""
        # Replace nan-values with maximum RSE
        if rse_av:
            np.nan_to_num(ODA._rse)
            rse_max = max(ODA._rse)
            np.where(ODA._rse == 0.0, rse_max, ODA._rse)

        # Calculate weight for each data set
            rse_min = min(ODA._rse)
            w = 1.0 - ((ODA._rse - rse_min) / rse_max)
        else:
            w = [1.0]*len(ODA._rse)
        refs = refs.join(pd.DataFrame(data={'weight': w}))
        return refs

    @classmethod
    def _add_weights_to_data(cls, data, refs):
        """adds weights from outlier detection to data"""
        for i in range(len(refs)):
            data.at[data['Original'] == refs.loc[i, 'Ref'], 'weight'] = refs.loc[i, 'weight']

    @classmethod
    def test_outlier_detection(cls, data):
        """tests whether outlier detection is reasonable for the given input data. This means that data are not
        detected as outliers only because they are in a different temperature region and more than one
        Arrhenius fit is required. Returns boolean, whether outlier detection should be applied"""
        refs = gf.DataManipulator().get_list_of_references(data)
        if len(refs) >= 3 and len(data) >= 3:
            return True
        else:
            return False

    @classmethod
    def od_main(cls, data: pd.DataFrame) -> pd.DataFrame:
        """Main function handling the outlier detection process."""
        ODA._test_data = pd.DataFrame()
        ODA._train_data = pd.DataFrame()
        ODA._rse = []
        ODA._obs = []
        ODA._M0 = []
        ODA._Q = []
        refs = gf.DataManipulator().get_list_of_references(data)
        rse_av = ODA._outlier_detection(data, refs)
        ODA._add_weights_to_data(data, ODA._get_weights(rse_av, refs))
        return data
