#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import collections

import pandas as pd

import general as gf
from assessment import errors as em
from assessment import fitting as fit
from db import ass_res as ar
from assessment.assessment_parameters import AssPar as ap


class InitTernary:
    """Initializes the assessment of ternary interaction parameters."""
    phase = None

    def __init__(self, phase):
        self._alloy = ap.alloy
        self.ass_obj_list = collections.OrderedDict()
        InitTernary.phase = phase

    def main(self):
        self._initialize()
        for key in self.ass_obj_list:
           self.ass_obj_list[key].ass_main()

    def _initialize(self):
        """Initializes all ternary alloys that can be assessed for the given alloy."""
        count = 0
        for el in self._alloy:
            for inter1 in self._alloy:
                if el != inter1:
                    for inter2 in self._alloy:
                        if inter2 != el and inter2 > inter1:
                            self.ass_obj_list[count] = TernInteraction(el, el, inter1, inter2)
                            count = count + 1
        for el in self._alloy:
            for inter1 in self._alloy:
                if el != inter1:
                    for inter2 in self._alloy:
                        if inter2 != el and inter2 > inter1:
                            for inter3 in self._alloy:
                                if inter3 != el and inter3 > inter2:
                                    self.ass_obj_list[count] = TernInteraction(el, inter1, inter2, inter3)
                                    count = count + 1


class TernInteraction:
    """Handles the assessment for ternary interaction parameters."""

    __slots__ = ('el', '_inter1', '_inter2', '_inter3', '_al', 'assessed', '_res', '_data_list', '_data', '_lower_ass_par')

    def __init__(self, el, inter1, inter2, inter3):
        self.el = gf.General.clean_input(el)
        self._inter1 = gf.General.clean_input(inter1)
        self._inter2 = gf.General.clean_input(inter2)
        self._inter3 = gf.General.clean_input(inter3)
        self._al = self._get_al()
        self.assessed = False
        self._res = pd.DataFrame(columns=ap.TERN_RES_HEADER)
        self._lower_ass_par = {}
        self._data_list = pd.DataFrame()
        self._data = pd.DataFrame()

    def _get_al(self):
        al = [self.el, self._inter1, self._inter2, self._inter3]
        al = list(set(al))
        return al

    def ass_main(self):
        """Decides whether an assessment is needed or assessed data can be loaded."""
        self._get_ass_data()
        if not self.assessed:
            self._assess()

    def _get_ass_data(self):
        """Returns the assessed data from the database for the given element in the given alloy."""
        res = ar.GetAssRes().get_tern()
        if not res.empty:
            res.drop(res[(res['DiffusingElement'] != self.el) | (res['IntEl1'] != self._inter1) |
                         (res['IntEl2'] != self._inter2)| (res['IntEl3'] != self._inter3) |
                         (res['Phase'] != InitTernary.phase)].index, inplace=True)
            res.reset_index(inplace=True, drop=True)
        if not res.empty:
            self.assessed = True
            res.rename(columns={'DiffusingElement': 'DiffEl'}, inplace=True)
            self._res = res
            self._store_results()

    def _assess(self):
        """Handles the assessment."""
        self._select_data()
        self._get_lower_data()
        if not self._data.empty:
            self._data.reset_index(inplace=True, drop=True)
            self._data_list = gf.PandasRelated.create_data_list(self._data)
            self._fit()
            self.assessed = True
            self._store_results()
        else:
            self._save_default()

    def _get_lower_data(self):
        """Collects the endmember assessment results and the assessment results from binary interactions."""
        un_dat = gf.AssResToDic.un_to_dict(ap.get_unary_results(), InitTernary.phase)
        bin_dat = gf.AssResToDic.bin_to_dict(ap.get_binary_results(), InitTernary.phase)
        self._lower_ass_par = dict(un_dat, **bin_dat)

    def _select_data(self):
        """Selects available data for binary assessment"""
        dat = ap.temp_data.copy()
        dat = gf.PandasRelated.select_data_by_diff_elem(self.el, dat)
        dat = gf.PandasRelated.select_data_by_alloy(self._al, dat)
        if len(self._al) == 4:
            dat = gf.PandasRelated.select_imp_data(self.el, dat)
        if not dat.empty:
            dat.reset_index(inplace=True, drop=True)
        self._data = dat

    def _fit(self):
        """Handles fitting of binary alloys"""
        result = gf.LnDfromAss(self.el, self._data_list, InitTernary.phase, self._lower_ass_par).main()
        basic_num_par = 2
        err = em.ErrorManager(basic_num_par, self._data_list['lnD'].tolist(), result.tolist())
        err.get_errors()
        res_temp = {'Mode': -1, 'MainEl':None, 'BIC': err.results['BIC'], 'AIC': err.results['AIC'], 'Value': 0.0}
        self._add_local_res(res_temp)
        self._one_parameter_fit(result)
        self._two_parameter_fit(result)
        self._three_parameter_fit(result)
        self._res = self._select_best_res()

    def _one_parameter_fit(self, result):
        """Fitting with one ternary parameter."""
        ### The result will be named by the main element that is chosen
        if len(self._data_list) > 2:
            for main_el in self._al:
                fitres = fit.FitTern(InitTernary.phase, self._data_list, result, diff_el=self.el, main_el1=main_el,
                                     int_el1=self._inter1,
                                   int_el2=self._inter2, int_el3=self._inter3).main()
                if fitres is not None:
                    ass_par_temp = dict(self._lower_ass_par, **fitres)
                    pred_bin = gf.LnDfromAss(self.el, self._data_list, InitTernary.phase, ass_par_temp).main()
                    num_par = 3
                    err = em.ErrorManager(num_par, self._data_list['lnD'].tolist(), pred_bin)
                    err.get_errors()
                    val = fitres["MQ_" + InitTernary.phase + "_" + self.el + "_in_" + self._inter1 + "_" + self._inter2 + "_" +
                                 self._inter3 + "_main_" + main_el + "_" + str(2)]
                    res_temp = {'Mode': main_el, 'MainEl': main_el, 'Value': val,  'BIC': err.results['BIC'], 'AIC': err.results['AIC']}
                    self._add_local_res(res_temp)

    def _two_parameter_fit(self, result):
        """Fitting with two ternary parameters."""
        # The result will be named by the two main elements
        if len(self._data_list) > 3:
            for main_el1 in self._al:
                for main_el2 in self._al:
                    if main_el2 > main_el1:
                        fitres = fit.FitTern(InitTernary.phase, self._data_list, result, diff_el=self.el, main_el1 = main_el1, main_el2 = main_el2, int_el1=self._inter1,
                            int_el2=self._inter2, int_el3=self._inter3).main()

                        if fitres is not None:
                            ass_par_temp = dict(self._lower_ass_par, **fitres)
                            pred_bin = gf.LnDfromAss(self.el, self._data_list, InitTernary.phase, ass_par_temp).main()
                            num_par = 4
                            err = em.ErrorManager(num_par, self._data_list['lnD'].tolist(), pred_bin)
                            err.get_errors()
                            val = fitres["MQ_" + InitTernary.phase + "_" + self.el + "_in_" + self._inter1 + "_" + self._inter2 + "_" +
                                self._inter3 + "_main_" + main_el1 + "_" + str(2)]
                            key = main_el1 + "_" + main_el2
                            res_temp = {'Mode': key, 'MainEl': main_el1, 'Value': val, 'BIC': err.results['BIC'],
                                       'AIC': err.results['AIC']}
                            self._add_local_res(res_temp)

                            val = fitres["MQ_" + InitTernary.phase + "_" + self.el + "_in_" + self._inter1 + "_" + self._inter2 + "_" +
                                self._inter3 + "_main_" + main_el2 + "_" + str(2)]
                            res_temp = {'Mode': key, 'MainEl': main_el2, 'Value': val, 'BIC': err.results['BIC'],
                                       'AIC': err.results['AIC']}
                            self._add_local_res(res_temp)

    def _three_parameter_fit(self, result):
        """Fitting with three ternary parameters."""
        # The result will be named with all
        if len(self._data_list) > 4:
            fitres = fit.FitTern(InitTernary.phase, self._data_list, result, diff_el=self.el, main_el1=self._inter1,
                             main_el2=self._inter2, main_el3=self._inter3, int_el1=self._inter1,
                            int_el2=self._inter2, int_el3=self._inter3).main()
            if fitres is not None:
                ass_par_temp = dict(self._lower_ass_par, **fitres)
                pred_bin = gf.LnDfromAss(self.el, self._data_list, InitTernary.phase, ass_par_temp).main()
                num_par = 5
                err = em.ErrorManager(num_par, self._data_list['lnD'].tolist(), pred_bin)
                err.get_errors()
                key = self._inter1 + "_" + self._inter2 + "_" + self._inter3
                val = fitres["MQ_" + InitTernary.phase + "_" + self.el + "_in_" + self._inter1 + "_" + self._inter2 + "_" +
                             self._inter3 + "_main_" + self._inter1 + "_" + str(2)]
                res_temp = {'Mode': key, 'MainEl': self._inter1, 'Value': val, 'BIC': err.results['BIC'],
                           'AIC': err.results['AIC']}
                self._add_local_res(res_temp)
                val = fitres["MQ_" + InitTernary.phase + "_" + self.el + "_in_" + self._inter1 + "_" + self._inter2 + "_" +
                             self._inter3 + "_main_" + self._inter2 + "_" + str(2)]
                res_temp = {'Mode': key, 'MainEl': self._inter2, 'Value': val, 'BIC': err.results['BIC'],
                           'AIC': err.results['AIC']}
                self._add_local_res(res_temp)
                val = fitres["MQ_" + InitTernary.phase + "_" + self.el + "_in_" + self._inter1 + "_" + self._inter2 + "_" +
                             self._inter3 + "_main_" + self._inter3 + "_" + str(2)]
                res_temp = {'Mode': key, 'MainEl': self._inter3, 'Value': val, 'BIC': err.results['BIC'],
                           'AIC': err.results['AIC']}
                self._add_local_res(res_temp)

    def _add_local_res(self, res):
        """Store the best parameters locally."""
        self._res = self._res.append({
            'Mode':res['Mode'], 'DiffEl': self.el, 'IntEl1': self._inter1, 'IntEl2': self._inter2,
            'IntEl3': self._inter3, 'MainEl':res['MainEl'], 'MQlowT': res['Value'], 'BP':None,
            'MQhighT':res['Value'], 'AIC': res['AIC'], 'BIC': res['BIC'], 'DP': len(self._data_list)}, ignore_index=True)

    def _store_results(self):
        """Adds assessment results of the given system to the local storage"""
        ap.add_ternary_result(self._res)

    def _select_best_res(self):
        """Selects the best result."""
        error = ap.SELECTION_CRITERION
        if len(self._res) > 1:
            min = self._res[error].min()
            if isinstance(min, list):
                min = min[0]
            res = self._res.loc[self._res[error] == min, :]
            res.reset_index(drop=True, inplace=True)
            return res
        else:
            return self._res

    def _save_default(self):
        """Saves default values if an assessment was not possible."""
        self._res = self._res.append({
            'Mode': "DEFAULT",'DiffEl': self.el, 'IntEl1': self._inter1, 'IntEl2': self._inter2, 'IntEl3': self._inter3,
            'MainEl': self.el, 'VMQlowT': 0.0, 'BP':None, 'MQhighT':0.0, 'AIC': 0.0, 'BIC': 0.0, 'DP': 0.0},
            ignore_index=True)
        self._res = self._res.append({
            'Mode': "DEFAULT", 'DiffEl': self.el, 'IntEl1': self._inter1, 'IntEl2': self._inter2, 'IntEl3': self._inter3,
            'MainEl': self._inter2, 'MQlowT': 0.0, 'BP': None, 'MQhighT': 0.0, 'AIC': 0.0,'BIC': 0.0, 'DP': 0.0},
            ignore_index=True)
        self._res = self._res.append({
            'Mode': "DEFAULT",'DiffEl': self.el, 'IntEl1': self._inter1, 'IntEl2': self._inter2, 'IntEl3': self._inter3,
            'MainEl': self._inter3, 'MQlowT': 0.0, 'BP':None, 'MQhighT':0.0, 'AIC': 0.0, 'BIC': 0.0, 'DP': 0.0},
            ignore_index=True)




