#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import pandas as pd

import general as gf


class SaveAssRes:
    """Save the assessment results in the SQL database."""

    def __init__(self, phase, un_res=pd.DataFrame(), bin_res=pd.DataFrame(), tern_res=pd.DataFrame()):
        self._phase = phase
        self._un_res = un_res
        self._bin_res = bin_res
        self._tern_res = tern_res

    def main(self):
        self._save_unary()
        self._save_binary()
        self._save_ternary()

    @staticmethod
    def _delete_old_results(conn):
        """Deletes the old results."""
        cur = conn.cursor()
        sql = """DELETE FROM DataInformation"""
        cur.execute(sql)
        conn.commit()


    def _add_gen_data(self, conn, gen_dat):
        """Adds the general data to the database."""
        cur = conn.cursor()
        placeholder = ", ".join(["?"] * len(gen_dat))
        sql = """INSERT INTO DataInformation VALUES ({values})""".format(values=placeholder)
        cur.execute(sql, gen_dat)
        conn.commit()

    def _save_unary(self):
        """Saves unary results in SQL."""
        with gf.General.create_connection(gf.General.PATH, "SQL_Databases/AssessmentResults/Endmember.db") as conn:
            SaveAssRes._delete_old_results(conn)
            if not self._un_res.empty:
                for i in range(len(self._un_res)):
                    el = self._un_res.loc[i, 'DiffEl']
                    mat = self._un_res.loc[i, 'Matrix']
                    gen_dat = (el, mat, self._phase, self._un_res.loc[i, 'MFlowT'],
                               self._un_res.loc[i, 'MQlowT'], self._un_res.loc[i, 'BP'],
                               self._un_res.loc[i, 'MFhighT'], self._un_res.loc[i, 'MQhighT'],
                               int(self._un_res.loc[i, 'DP']), self._un_res.loc[i, 'Mode'])
                    self._add_gen_data(conn, gen_dat)

    def _save_binary(self):
        """Saves binary results in SQL."""
        conn = gf.General.create_connection(gf.General.PATH, "SQL_Databases/AssessmentResults/BinaryInteractions.db")
        SaveAssRes._delete_old_results(conn)
        if not self._bin_res.empty:
            res = self._bin_res
            for i in range(len(res)):
                el = res.loc[i, 'DiffEl']
                inter_elem1 = res.loc[i, 'IntEl1']
                inter_elem2 = res.loc[i, 'IntEl2']
                ho = int(res.loc[i, 'HighestOrder'])
                order = int(res.loc[i, 'Order'])
                val_lowT = res.loc[i, 'MQlowT']
                val_highT = res.loc[i, 'MQhighT']
                bp = res.loc[i, 'BP']
                dp = int(res.loc[i, 'DP'])
                gen_dat = (self._phase, el, inter_elem1, inter_elem2, ho, order, val_lowT, bp,
                           val_highT, dp)
                self._add_gen_data(conn, gen_dat)
            conn.close()

    def _save_ternary(self):
        """Saves ternary result in SQL."""
        conn = gf.General.create_connection(gf.General.PATH, "SQL_Databases/AssessmentResults/TernaryInteractions.db")
        SaveAssRes._delete_old_results(conn)
        if not self._tern_res.empty:
            res = self._tern_res
            for i in range(len(res)):
                el = res.loc[i, 'DiffEl']
                inter_elem1 = res.loc[i, 'IntEl1']
                inter_elem2 = res.loc[i, 'IntEl2']
                inter_elem3 = res.loc[i, 'IntEl3']
                main_elem = res.loc[i, 'MainEl']
                val_lowT = res.loc[i, 'MQlowT']
                bp = res.loc[i, 'BP']
                val_highT = res.loc[i, 'MQhighT']
                dp = res.loc[i, 'DP']
                gen_dat = (self._phase, el, inter_elem1, inter_elem2, inter_elem3, main_elem, val_lowT, bp, val_highT,dp)
                self._add_gen_data(conn, gen_dat)
            conn.close()


class GetAssRes:
    """Collects the assessment results from the SQL databases"""

    def get_un(self):
        """Selects the endmember assessment results from the database."""
        file = 'SQL_Databases/AssessmentResults/Endmember.db'
        conn = gf.General.create_connection(gf.General.PATH, file)
        sql_command = """SELECT * FROM DataInformation; """
        data = pd.read_sql_query(sql_command, conn)
        conn.close()
        data.rename(columns={'breakpoint': 'BP', 'DataPoints': 'DP'}, inplace=True)
        return data

    def get_bin(self):
        """Selects the assessed binary interaction terms from the database."""
        file = 'SQL_Databases/AssessmentResults/BinaryInteractions.db'
        conn = gf.General.create_connection(gf.General.PATH, file)
        sql_command = """SELECT * FROM DataInformation; """
        data = pd.read_sql_query(sql_command, conn)
        conn.close()
        data.rename(columns={
            'breakpoint': 'BP', 'DataPoints': 'DP', 'Interaction Element1': 'IntEl1', 'Interaction Element2': 'IntEl2',
            'order': 'Order'}, inplace=True)
        return data

    def get_tern(self):
        """Selects the assessed ternary interaction terms from the database."""
        file = 'SQL_Databases/AssessmentResults/TernaryInteractions.db'
        conn = gf.General.create_connection(gf.General.PATH, file)
        sql_command = """SELECT * FROM DataInformation; """
        data = pd.read_sql_query(sql_command, conn)
        conn.close()
        data.rename(columns={
            'breakpoint': 'BP', 'DataPoints': 'DP', 'Interaction Element1': 'IntEl1', 'Interaction Element2': 'IntEl2',
            'Interaction Element3': 'IntEl3', 'Main Element': 'MainEl'}, inplace=True)
        return data

