#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

from sqlite3 import Error

import general as gf
from general import InitRes as rm
import sqlite3


class InitDB:

    __slots__ = '_path'

    def __init__(self):
        self._path = rm.get_path() + "/SQL_Databases"

    def main(self):
        """Initializes the databases if they do not exist yet."""
        self._create_directories()
        self._init_raw_data()
        self._init_reference_db()
        self._init_results()

    def _create_directories(self):
        gf.General.create_dir(self._path)
        gf.General.create_dir(self._path + "/AssessmentResults")
        gf.General.create_dir(self._path + "/RawData")

    def _init_results(self):
        # self._init_tracer_db()
        self._init_unary_res()
        self._init_binary_res()
        self._init_ternary_res()

    def _init_reference_db(self):
        """Initializes database which stores all references."""
        file = self._path + "/RawData/References.db"
        with sqlite3.connect(file) as conn:
            self._create_ref_table(conn)

    def _init_raw_data(self):
        """Initializes tracer database."""
        file = self._path + "/RawData/Tracer.db"
        with sqlite3.connect(file) as conn:
            self._create_raw_data_table(conn)
            self._create_table_comp(conn)

    def _init_unary_res(self):
        """Initializes database about unary assessments."""
        file = self._path + "/AssessmentResults/Endmember.db"
        with sqlite3.connect(file) as conn:
            self._create_unary_table(conn)

    def _init_binary_res(self):
        """Initializes databases that stores the excess parameters"""
        file = self._path + "/AssessmentResults/BinaryInteractions.db"
        with sqlite3.connect(file) as conn:
            self._create_bin_table(conn)

    def _init_ternary_res(self):
        """Initializes databases that stores the excess parameters"""
        file = self._path + "/AssessmentResults/TernaryInteractions.db"
        with sqlite3.connect(file) as conn:
            self._create_tern_table(conn)

    def _create_ref_table(self, conn: sqlite3.Connection):
        """Creates the table for the references"""
        sql_table = """ CREATE TABLE IF NOT EXISTS Ref ("ID" text, "Type" text, "Title" text,
         "First author (last name)" text, "Second author (last name)" text, "Third author (last name)" text, "Year" float,
         "Journal/publisher" text, "Volume" float); """
        try:
            c = conn.cursor()
            c.execute(sql_table)
        except Error as e:
            print(e)

    def _create_table_comp(self, conn: sqlite3.Connection):
        """ create a table for the composition of the tracer data"""
        sql_table = """ CREATE TABLE IF NOT EXISTS Composition (ID text, AC float, AG float, AL float, AR float, "ARS" float, 
        "AT" float, AU float, B float, BA float, BE float, BH float, BI float, C float, CA float, CD float, CE float, CN float, 
        CO float, CR float, CS float, CU float, DB float, DS float, ER float, EU float, F float, FE float, FR float, GA float, 
        GE float, GD float, H float, HE float, HF float, HG float, 'HO' float, HS float, "IND" float, IR text, K float, KR float, 
        LA float, LI float, LU float, N float, NA float, MG float, MN float, MO text, MT float,  NB float, ND float, NE float, 
        NI float, O float, OS float, P float, PB float, PD float, PM float, PO text, PT float, PU float, RA float, RB float, 
        RE float, RF float, RG float, RH float, RN float, RU float, S float, SB text, SC float, SE float, SG float, SI float, 
        SM float, SN float, SR float,  TA float,  TC float, TE float, TH float, TI float, TL float, TM float, U float, V float, 
        W float,XE float, Y float, YB float, ZN float,ZR float); """
        try:
            c = conn.cursor()
            c.execute(sql_table)
        except Error as e:
            print(e)

    def _create_raw_data_table(self, conn: sqlite3.Connection):
        """ create a table with all information about the tracer data except the composition """
        sql_table = """ CREATE TABLE IF NOT EXISTS DataInformation (ID text, "DiffusingElement" text, Tmin float, Tmax 
        float, Phase text, D float, "D0" float, "Q" float, 'Data Points' float, "Taken from" text, "Original" text, 
        "Taken from Plot" BIT, Details text); """
        try:
            c = conn.cursor()
            c.execute(sql_table)
        except Error as e:
            print(e)

    def _create_unary_table(self, conn: sqlite3.Connection):
        """ create a table with all information about the Arrhenius Parameter data except the composition """
        # Info: BIT datatype in SQL can represent a boolean, since it can be either 0 or 1; 0 means false and 1 means
        # true
        sql_table = """ CREATE TABLE IF NOT EXISTS DataInformation (DiffusingElement text, Matrix text,
        Phase text, MFlowT float, MQlowT float, breakpoint, MFhighT float, MQhighT float, "DataPoints" int, 
        "Mode" text); """
        try:
            c = conn.cursor()
            c.execute(sql_table)
        except Error as e:
            print(e)

    def _create_bin_table(self, conn: sqlite3.Connection):
        """ create a table with all information about binary and ternary assessment results. Parameter should be A if it
         is an excess parameter for the traditional fitting with one parameter. If the PW fitting is used it should be
         M for M0 excess and Q for Q0 excess. """
        sql_table = """ CREATE TABLE IF NOT EXISTS DataInformation ("Phase" text, 
        "DiffusingElement" text, "Interaction Element1" text, "Interaction Element2" text, "HighestOrder" int, "order" int, "MQlowT" float, "breakpoint" float, "MQhighT" float, 
        "DataPoints" int); """
        try:
            c = conn.cursor()
            c.execute(sql_table)
        except Error as e:
            print(e)

    def _create_tern_table(self, conn: sqlite3.Connection):
        """ create a table with all information about binary and ternary assessment results. Parameter should be A if it
         is an excess parameter for the traditional fitting with one parameter. If the PW fitting is used it should be
         M for M0 excess and Q for Q0 excess. """
        sql_table = """ CREATE TABLE IF NOT EXISTS DataInformation ("Phase" text,
        "DiffusingElement" text, "Interaction Element1" text, "Interaction Element2" text, "Interaction Element3" text,
        "Main Element" text, "MQlowT" float, "breakpoint" float, "MQhighT" float, "DataPoints" int); """
        try:
            c = conn.cursor()
            c.execute(sql_table)
        except Error as e:
            print(e)
