#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import general as gf
import os
import pandas as pd
import numpy as np


class RawDataAdder:
    """Adds raw data to the SQL database"""

    __slots__ = ('_comp_data', '_gen_data', '_file', '_ID')

    def __init__(self, comp_data: list, gen_data:pd.DataFrame):
        self._comp_data = comp_data
        self._gen_data = gen_data
        self._file = gf.General.PATH + "/SQL_Databases/RawData/Tracer.db"

    def main(self):
        with gf.General.create_connection(None, self._file) as conn:
            self._clean_old_data(conn)
            #self._get_ID(conn)
            self._add_gen_data(conn)
            self._add_comp(conn)

    def _clean_old_data(self, conn):
        cur = conn.cursor()
        sql_command = """DELETE FROM DataInformation"""
        cur.execute(sql_command)
        cur = conn.cursor()
        sql_command = """DELETE FROM Composition"""
        cur.execute(sql_command)

    def _add_comp(self, conn):
        """Adds the composition to the Database."""
        id = 0
        for comp in self._comp_data:
            comp['ID'] = id
            cur = conn.cursor()
            placeholder = ", ".join(["?"] * (len(comp)))
            columns = ",".join(comp.keys())
            stmt = "insert into Composition ({columns}) values ({values});".format(columns=columns,
                                                                                   values=placeholder)
            cur.execute(stmt,list(comp.values()))
            conn.commit()
            id += 1

    def _add_gen_data(self, conn):
        """Adds the general data to the database."""
        id = range(len(self._gen_data))
        self._gen_data['ID'] = id
        self._gen_data.to_sql('DataInformation', conn, if_exists='append', index=False)

    @classmethod
    def readd_data(cls, comp_data:pd.DataFrame, gen_data:pd.DataFrame):
        file = gf.General.PATH + "/SQL_Databases/RawData/Tracer.db"
        with gf.General.create_connection(None, file) as conn:
            gen_data.to_sql("DataInformation", conn,  if_exists='replace', index=False)
            comp_data.to_sql("Composition", conn,  if_exists='replace', index=False)

class RefAdder:
    """Handles the addition of new references to the SQL database"""

    def __init__(self, data: pd.DataFrame):
        self._data = data
        self._file = os.getcwd() + "/SQL_Databases/RawData/References.db"

    def main(self):
        with gf.General.create_connection(None, self._file) as conn:
            cur = conn.cursor()
            sql_command = """DELETE FROM Ref"""
            cur.execute(sql_command)
            self._data.to_sql("Ref", conn,  if_exists='append', index=False)


class GetData:
    """Collects raw data from the database."""

    __slots__ = ('_data', '_file')

    def __init__(self):
        self._data = pd.DataFrame()
        self._file = gf.General.PATH + "/SQL_Databases/RawData/Tracer.db"

    def main(self) -> pd.DataFrame:
        with gf.General.create_connection(None, self._file) as conn:
            rows_gen = pd.read_sql("""SELECT * FROM DataInformation""", conn)
            comp = pd.read_sql("""SELECT * FROM Composition""", conn)
            res = pd.concat([rows_gen, comp],axis=1, sort=False)
            res.drop_duplicates(inplace=True)
            res = res.loc[:, ~res.columns.duplicated()]
            res.replace(to_replace=["Null", "NULL", "null", None], value=np.nan, inplace=True, regex=True)
            res.reset_index(inplace=True, drop=True)
            res.drop(['ID'],axis=1,  inplace=True)
            res.rename(columns={'Data Points': "DP"}, inplace=True)
        return res


class GetRef:
    """Collects references from the database."""

    __slots__ = ('_data', '_file')

    def __init__(self):
        self._data = pd.DataFrame()
        self._file = gf.General.PATH + "/SQL_Databases/RawData/References.db"

    def main(self) -> pd.DataFrame:
        with gf.General.create_connection(None, self._file) as conn:
            res = pd.read_sql("""SELECT * FROM Ref""", conn)
            res.replace(to_replace=["Null", "NULL", "null", None], value=np.nan, inplace=True, regex=True)
            res.reset_index(inplace=True, drop=True)
        return res

