#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import os
import re
import sqlite3
from sqlite3 import Error

import numpy as np
import pandas as pd

from setup import Setup as setup


class General:
    """This class contains parameters and functions for general purposes."""
    PATH = os.getcwd()

    @classmethod
    def create_dir(cls, dicname: str):
        """Creates new directories"""
        try:
            os.mkdir(dicname)
        except OSError:
            pass
            if setup.DEBUG:
                print("Creation of the directory %s failed" % dicname)
        else:
            if setup.DEBUG:
                print("Successfully created the directory %s" % dicname)

    @classmethod
    def create_connection(cls, path: str, filename:str) -> sqlite3.Connection:
        """ Create a database connection to a SQLite database. """
        file = filename
        if path is not None:
            file = path + "/" + filename
        try:
            conn = sqlite3.connect(file)
        except Error as e:
            print(e)
        finally:
            return conn

    @classmethod
    def get_file_modifier(cls, path: str, filename:str) -> str:
        """Check if a file exists and return the appropriate file modifier."""
        if os.path.exists(path + filename):
            return 'a'  # append if already exists
        else:
            return 'w+'  # make a new file if not

    @classmethod
    def clean_input(cls, inp):
        """Returns the input with upper case letters and withput whitespaces around."""
        if not isinstance(inp, str):
            res = [x.strip() for x in inp]
            res = [x.upper() for x in res]
        else:
            res = inp.upper().strip()
        return res


class PhysicalProperties:
    """Stores available physical properties like melting points for the different elements in SI-Units."""
    _MELTING_POINTS = {"LI": 453.65, "BE": 1560.15, "B": 2348.15, "N2": 63.15, "O2": 54.36, "NE": 24.56, "NA": 370.944,
                       "MG": 923.15, "AL": 933.47, "SI": 1687.15, "Cl2": 171.65,  "AR": 83.81, "K": 336.65, "CA": 1115.15,
                       "SC": 1814.15, "TI": 1941.15, "V": 2183.15, "CR": 2180.15, "MN": 1519.15, "FE": 1811.15,
                       "CO": 1768.15, "NI": 1728.15, "CU": 1357.77, "ZN": 692.68, "GA": 302.921, "GE": 1211.4,
                       "AS": 1090.15, "SE": 453.15, "BR2": 265.95, "KR": 115.77, "RB": 312.45, "SR": 1050.15,
                       "Y": 1795.15, "ZR": 2128.15, "NB": 2750.15, "MO": 2896.15, "TC": 2430.15, "RU": 2607.15,
                       "RH": 2237.15, "PD": 1828.05, "AG": 1234.93, "CD": 594.22, "IN": 429.75, "SN": 505.08,
                       "SB": 903.78, "TE": 722.66, "I2": 386.85, "XE": 161.355, "CS": 301.65, "BA": 1000.15,
                       "LA": 1191.15, "CE": 1071.15, "PR": 1204.15, "ND": 1294.15, "PM": 1315.15, "SM": 1347.15,
                       "EU": 1095.15, "GD": 1586.15, "TB": 1629.15, "DY": 1685.15, "HO": 1747.15, "ER": 1802.15,
                       "TM": 1818.15, "YB": 1092.15, "LU": 1936.15, "HF": 2506.15, "TA": 3290.15,
                       "W": 3695.15, "RE": 3459.15, "OS": 3306.15, "IR": 2719.15, "PT": 2041.55, "AU": 1337.33,
                       "HG": 234.313, "TL": 577.15, "PB": 600.61, "BI": 544.55, "PO": 527.15, "AT": 575.15,
                       "RN": 202.15, "FR": 300.15, "RA": 973.15, "AC": 1324.15, "TH": 2023.15, "PA": 1841.15,
                       "U": 1408.15, "NP": 917.15, "PU": 913.15, "AM": 1449.15, "CM": 1618.15, "ES": 1133.15,
                       "MD": 1100.15, "NO": 1100.15, "LR": 1900.15}

    _CURIE_TEMP = {"FE": 1043, "CO": 1400, "NI": 627, "GD": 292, "DY": 88, "MNBI": 630, "MNSB": 587, "CRO2": 386,
                   "MNAS": 318, "EUO": 68, "FE2O3": 948, "FEOFE2O3": 858}

    _PHASE_SYNONYMS = {"FCC": "FCC_A1", "FCC_A1": "FCC"}

    @classmethod
    def get_phase_synonym(cls, phase: str):
        """Checks if there are synonyms for a given phase."""
        if phase in PhysicalProperties._PHASE_SYNONYMS:
            return PhysicalProperties._PHASE_SYNONYMS[phase]
        else:
            return None

    @classmethod
    def get_mp(cls, el: str) -> float:
        """Returns the melting point of a given element."""
        if el in PhysicalProperties._MELTING_POINTS:
            return PhysicalProperties._MELTING_POINTS[el]
        else:
            return None

    @classmethod
    def get_curie_temp(cls, el: str) -> float:
        """Returns the Curie temperature of a given element."""
        if el in PhysicalProperties._CURIE_TEMP:
            return PhysicalProperties._CURIE_TEMP[el]
        else:
            return None


class PandasRelated:
    """Contains funtions modifying Pandas data frames and are used in several classes."""

    @classmethod
    def select_data_by_diff_elem(cls, el:str, dat:pd.DataFrame) -> pd.DataFrame:
        """Selects data from a dataframe for a given diffusing element."""
        if setup.DEBUG:
            pd.set_option('mode.chained_assignment', 'raise')
        data = dat.copy()
        if not data.empty:
            data.drop(data[data['DiffusingElement'] != el].index, inplace=True)
            data.reset_index(drop=True, inplace=True)
        return data

    @classmethod
    def select_data_by_alloy(cls, al: list, dat: pd.DataFrame) -> pd.DataFrame:
        """Selects data from a dataframe for the given alloy."""
        if setup.DEBUG:
            pd.set_option('mode.chained_assignment', 'raise')
        data = dat.copy()
        amount_nan = len(setup.TOTAL_EL) - len(al)
        if not data.empty:
            data = data.loc[data[setup.TOTAL_EL].isnull().sum(axis=1) == amount_nan].copy()
            data = data.loc[data[al].isnull().sum(axis=1) == 0].copy()
            data.reset_index(inplace=True, drop=True)
        return data

    @classmethod
    def select_imp_data(cls, el: str, dat: pd.DataFrame) -> pd.DataFrame:
        """Select data for impurity diffusion of element el."""
        if setup.DEBUG:
            pd.set_option('mode.chained_assignment', 'raise')
        data = dat.copy()
        amount_nan = len(setup.TOTAL_EL) - 2
        data = data.loc[data.isnull().sum(axis=1) >= amount_nan]
        data.reset_index(inplace=True, drop=True)
        data[setup.TOTAL_EL] = data[setup.TOTAL_EL].astype(float)
        data = data.loc[data[[el]].isin([0.0]).sum(axis=1) != 0]
        data.to_csv('test.csv')
        data.reset_index(inplace=True, drop=True)
        return data

    @classmethod
    def select_self_diff_data(cls, el:str, dat:pd.DataFrame) -> pd.DataFrame:
        """Select data for self-diffusion."""
        data = dat.copy()
        if not data.empty:
            data[el] = data[el].astype(float)
            data = data[~data[el].isin([0.0])]
            data.reset_index(inplace=True, drop=True)
        return data

    @classmethod
    def select_only_bin_data(cls, dat: pd.DataFrame) -> pd.DataFrame:
        """Select only data for binary diffusion."""
        data = dat.copy()
        amount_nan = len(setup.ELEMENTS) - 2
        if not data.empty:
            data = data.loc[data.isnull().sum(axis=1) == amount_nan]
            data.reset_index(inplace=True, drop=True)
            data[setup.TOTAL_EL] = data[setup.TOTAL_EL].astype(float)
            data = data.loc[data[setup.TOTAL_EL].isin([0.0, 1.0]).sum(axis=1) == 0]
            data.reset_index(inplace=True, drop=True)
        return data

    @classmethod
    def get_el_list_from_al_list(cls, al_list: list) -> list:
        """Create a list of elements from all the elements occuring in the alloys. """
        el_l =[]
        for al in al_list:
            for el in al:
                el_l.append(el)
        el_l = list(set(el_l))
        return el_l

    @classmethod
    def get_alloy_list(cls, data):
        """Create a list of alloys from the dataframe."""
        al = []
        if not data.empty:
            als = data[setup.TOTAL_EL].copy()
            als.to_csv('test.csv')
            als.dropna(axis=1, how='all', inplace=True)
            als.drop_duplicates(inplace=True)
            als.reset_index(inplace=True, drop=True)
            for i in range(len(als)):
                temp_al = als.iloc[i].copy()
                temp_al.dropna(inplace=True)
                al.append(temp_al.index.tolist())
            return al
        else:
            return []

    @classmethod
    def create_data_list(cls, data: pd.DataFrame) -> pd.DataFrame:
        """Create a data list that can be used for the fitting procedure."""
        el_l = PandasRelated.get_el_list_from_al_list(PandasRelated.get_alloy_list(data.copy()))
        col = el_l
        col.append('lnD')
        col.append('Temp')
        col.append('DiffusingElement')
        data_list = data[col].copy()
        data_list.fillna(0.0, inplace=True)
        data_list.reset_index(inplace=True, drop=True)
        return data_list

    @classmethod
    def create_bin_help_data_list(cls, el_l, tmin, tmax, cmin=0.0, cmax=1.0, n=100):
        """Create data list for a unary that has to be assessed from binary data?"""
        t = np.linspace(tmin, tmax, n)
        if tmin == tmax:
            t = [tmin]
        c = np.linspace(cmin, cmax, n)
        if cmin == cmax:
            c = [cmin]
        df = pd.DataFrame(columns=el_l)
        df_help = pd.DataFrame(columns=el_l)
        df_help[el_l[0]] = c
        df_help[el_l[1]] = 1 - df_help[el_l[0]]
        for temp in t:
            df_help['Temp'] = pd.Series(n*[temp])
            if df.empty:
                df = df_help.copy()
            else:
                df = pd.concat([df, df_help], ignore_index=True)
        return df


class Conv:
    """Calculates atomic mobility"""
    R = 8.314

    def arr_par_into_d(self, data: pd.DataFrame) -> pd.DataFrame:
        """Calculates from Arrhenius parameters the D values and returns a dataframe only containing single point
        diffusivities."""
        if setup.DEBUG:
            pd.set_option('mode.chained_assignment', 'raise')
        data.reset_index(drop=True, inplace=True)
        arr_par_dat = data.loc[data['D'].isnull()].copy()
        data.drop(data[data['D'].isnull()].index, inplace=True)
        if not data.empty:
            data.reset_index(inplace=True, drop=True)
        arr_par_dat = self._calc_arr_par(arr_par_dat)
        result = self._polish_result(data, arr_par_dat)
        return result

    def _prepare_arr_par_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Prepare data to replace Arrhenius Parameters."""
        data.drop(data[(data['D0'].isnull()) | (data['Q'].isnull()) | (data['Phase'].isnull())].index, inplace=True)
        data.drop(data[(data['Tmin'].isnull()) & (data['Tmax'].isnull())].index, inplace=True)
        data.reset_index(inplace=True, drop=True)
        data[['DP']] = data[['DP']].replace(to_replace=[np.nan, 0, 1, 0.0, 1.0], value=2)
        data['Tmin'].fillna((data['Tmax'] - 10.0), inplace=True)
        data['Tmax'].fillna((data['Tmin'] + 10.0), inplace=True)
        data['Tdiff'] = data['Tmax'] - data['Tmin']
        data.loc[data['Tdiff'] == 0.0, 'Tmin'] = data['Tmin'] - 10.0
        data.loc[data['Tdiff'] == 0.0, 'Tmax'] = data['Tmin'] + 20.0
        data['ID'] = range(len(data))
        return data

    def _calc_arr_par(self, arr_par_dat: pd.DataFrame) -> pd.DataFrame:
        if not arr_par_dat.empty:
            arr_par_dat = self._prepare_arr_par_data(arr_par_dat)
            ID_list = arr_par_dat['ID'].tolist()
            arr_par_dat = arr_par_dat.loc[arr_par_dat.index.repeat(arr_par_dat['DP'])].reset_index(drop=True)
            arr_par_dat.reset_index(inplace=True, drop=True)
            tot_ind = 0
            for j, ind in enumerate(ID_list):
                amount = len(arr_par_dat.loc[arr_par_dat['ID'] == ind])
                row = arr_par_dat.loc[arr_par_dat['ID'] == ind].iloc[0]
                tmin = row['Tmin']
                tmax = row['Tmax']
                d0 = float(row['D0'])
                q = float(row['Q'])
                temp_list = np.linspace(float(tmin), float(tmax), amount)
                d_list = float(d0) * np.exp(-float(q) / (Conv.R * temp_list))
                for i in range(amount):
                    arr_par_dat.at[tot_ind+i, 'D'] = d_list[i]
                    arr_par_dat.at[tot_ind+i, 'Tmin'] = temp_list[i]
                    arr_par_dat.at[tot_ind+i, 'Tmax'] = temp_list[i]
                tot_ind += amount
        return arr_par_dat

    def _polish_result(self, data: pd.DataFrame, arr_par_dat:pd.DataFrame) -> pd.DataFrame:
        result = pd.concat([data, arr_par_dat], sort=True)
        result.rename(columns={"Tmin": "Temp"}, inplace=True)
        result.reset_index(drop=True, inplace=True)
        result['D'] = pd.to_numeric(result.loc[:, 'D'])
        result['Temp'] = pd.to_numeric(result.loc[:,'Temp'])
        result.drop(['Tmax', 'Tdiff', 'D0', 'Q', 'Details', 'DP', 'Taken from', 'Taken from Plot'], axis=1, inplace=True)
        return result


class LnDfromAss:
    """Calculate LnD values from assessment results for given concentrations and temperatures."""
    R = 8.314

    def __init__(self, diff_el,  data_list, phase, ass_par, own_data=True):
        self._data_list = data_list #contains compositions and temperature, same structure as in fitting
        self._ass_par = ass_par
        self._pred_list = pd.Series()
        self._phase = phase
        self._el_l = None
        self._own_data = own_data
        self._diff_el = diff_el

    def main(self):
        """Calculate LnD values from assessment results for given concentrations and temperatures."""
        temp = self._data_list['Temp']
        comp = self._data_list.copy()
        comp.drop(['Temp'], axis=1, inplace=True)
        if 'DiffusingElement' in self._data_list.columns:
            comp.drop(['DiffusingElement'], axis=1, inplace=True)
        if 'lnD' in self._data_list.columns:
            comp.drop(['lnD'], axis=1, inplace=True)
        self._el_l = list(comp.columns)
        self._get_linear_contribution(temp, comp)
        self._get_binary_interactions(temp, comp)
        self._get_ternary_interactions(temp, comp)
        return self._pred_list

    def _get_linear_contribution(self, temp, comp):
        """Calculate linear contribution."""
        for mat in self._el_l:
                try:
                    mf= self._ass_par["MF_" + self._phase + "_" + self._diff_el + "_in_" + mat + "_-1"]
                except KeyError:
                    mf = 1
                try:
                    mq= self._ass_par["MQ_" + self._phase + "_" + self._diff_el + "_in_" + mat + "_-1"]
                except KeyError:
                    mq= 0

                bp = None
                if isinstance(mf, dict):
                    mf_lowT = mf['Value_lowT']
                    mf_highT = mf['Value_highT']
                else:
                    mf_lowT = float(mf)
                    mf_highT = float(mf)
                if isinstance(mq, dict):
                    bp = mq['BP']
                    mq_lowT = mq['Value_lowT']
                    mq_highT = mq['Value_highT']
                else:
                    mq_lowT = float(mq)
                    mq_highT = float(mq)

                c = comp[mat].astype(float)
                if bp is not None:
                    mf = []
                    mq = []
                    for i in range(len(temp)):
                        if temp[i] > bp:
                            mf.append(mf_highT)
                            mq.append(mq_highT)
                        else:
                            mf.append(mf_lowT)
                            mq.append(mq_lowT)
                else:
                    mf = mf_lowT
                    mq = mq_lowT

                if not self._own_data:
                    #f = (np.log(mf) + mq) / (LnDfromAssOldAssRes.R * temp)
                    f = c* (mq / (LnDfromAss.R * temp))

                else:
                    f = c*((LnDfromAss.R * temp * np.log(mf) - mq) / (LnDfromAss.R * temp))

                if 'DiffEl' in self._data_list.columns:
                    for i in range(len(self._data_list['DiffEl'].tolist())):
                        if self._data_list.loc[i, 'DiffEl'] != self._diff_el:
                            f[i] = 0

                if self._pred_list.empty:
                    self._pred_list = f
                else:
                    self._pred_list = self._pred_list + f

    def _get_binary_interactions(self, temp, comp):
        """Calculate binary contribution."""
        col_list = comp.columns.values
        for key in self._ass_par:
            bin = False
            t = re.split("_", key)
            int_el1 = "Null"
            int_el2 = "Null"
            el = "Null"
            ord = -1
            if len(t) == 8:
                el = t[3]
                ord = int(t[7])
                int_el1 = t[5]
                int_el2 = t[6]
                bin = True
            elif len(t) == 7:
                bin = True
                el = t[2]
                ord = int(t[6])
                int_el1 = t[4]
                int_el2 = t[5]
            if bin:
                if int_el1 in col_list and int_el2 in col_list and ord != -1 and ord != 2 and ord != 4 and el == self._diff_el:
                    if isinstance(self._ass_par[key], dict):
                        p = float(self._ass_par[key]['Value_lowT'])
                    else:
                        p = float(self._ass_par[key])
                    c1 = comp[int_el1].astype(float)
                    c2 = comp[int_el2].astype(float)
                    self._pred_list += c1 * c2 * p * (c1 - c2) ** ord / (LnDfromAss.R * temp)

    def _get_ternary_interactions(self, temp, comp):
        """Calculate ternary contributions."""
        col_list = comp.columns.values
        for key in self._ass_par:
            tern = False
            t = re.split("_", key)
            int_el1 = "Null"
            int_el2 = "Null"
            int_el3 = "Null"
            main_el = "Null"
            ord ="Null"
            el = "Null"
            if len(t) == 10:
                el = t[2]
                int_el1 = t[4]
                int_el2 = t[5]
                int_el3 = t[6]
                main_el = t[8]
                ord = int(t[9])
                tern = True
            elif len(t) == 11:
                tern = True
                el = t[3]
                int_el1 = t[5]
                int_el2 = t[6]
                int_el3 = t[7]
                main_el = t[9]
                ord = int(t[10])
            if tern:
                if int_el1 in col_list and int_el2 in col_list and int_el3 in col_list and main_el in col_list and int(ord) == 2 and el == self._diff_el:
                    if isinstance(self._ass_par[key], dict):
                        p = float(self._ass_par[key]['Value_lowT'])
                    else:
                        p = float(self._ass_par[key])
                    c1 = comp[int_el1].astype(float)
                    c2 = comp[int_el2].astype(float)
                    c3 = comp[int_el3].astype(float)
                    c_main = comp[main_el].astype(float)
                    mu = c_main + (1-c1-c2-c3)/3
                    self._pred_list += c1 * c2 * c3 * mu * p / (LnDfromAss.R * temp)


class DataManipulator:
    """Adds necessary values to the data, e.g. calculates lnD from D."""

    @classmethod
    def _add_lnD(cls, data: pd.DataFrame) -> pd.DataFrame:
        """Adds lnD to all data."""
        if not 'lnD' in list(data.columns):
            lnD = pd.DataFrame({'lnD': np.log(data['D'])})
            return data.join(lnD)
        else:
            return data

    @classmethod
    def add_invT(cls, data: pd.DataFrame) -> pd.DataFrame:
        """Adds inverse Temperature to all data"""
        if not 'invT' in list(data.columns):
            invT = pd.DataFrame({'invT': 1.0/data['Temp']})
            return data.join(invT)
        else:
            return data

    @classmethod
    def prepare_data(cls, data: pd.DataFrame) -> pd.DataFrame:
        """Prepare data for assessment"""
        if not data.empty:
            data = DataManipulator.add_invT(data)
            data = DataManipulator._add_lnD(data)
        return data

    @classmethod
    def get_list_of_references(cls, data: pd.DataFrame):
        """Prepares a list of references for these data"""
        ref_list = pd.DataFrame(columns=['Ref', 'DP'])
        if not data.empty:
            references = data['Original'].tolist()
            unique_refs = list(set(references))
            dps = []
            for ref in unique_refs:
                dps.append(len(data.loc[data['Original']==ref]))
            ref_list['Ref'] = unique_refs
            ref_list['DP'] = dps
        return ref_list


class AssResToDic:
    """Transform assessment results into another storage for further reuse."""
    tmin = 300
    tmax = 6000

    @classmethod
    def un_to_dict(cls, asspar, phase):
        """Save unary results in dictionaries."""
        asspar.reset_index(inplace=True, drop=True)
        d = {}
        for i in range(len(asspar)):
            el = asspar.loc[i, 'DiffEl']
            mat = asspar.loc[i, 'Matrix']
            k_mq = "MQ_" + phase + "_" + el + "_in_" + mat + "_" + str(-1)
            d[k_mq] = {}
            d[k_mq]['Value_lowT'] = asspar.loc[i, 'MQlowT']
            d[k_mq]['Value_highT'] = asspar.loc[i, 'MQhighT']
            d[k_mq]['BP'] = asspar.loc[i, 'BP']
            k_mf = "MF_" + phase + "_" + el + "_in_" + mat + "_" + str(-1)
            d[k_mf] = {}
            d[k_mf]['Value_lowT'] = asspar.loc[i, 'MFlowT']
            d[k_mf]['Value_highT'] = asspar.loc[i, 'MFhighT']
            d[k_mf]['BP'] = asspar.loc[i, 'BP']
        return d

    @classmethod
    def bin_to_dict(cls, asspar, phase):
        """Save binary results in dictionaries."""
        asspar.reset_index(inplace=True, drop=True)
        d = {}
        for i in range(len(asspar)):
            order = int(asspar.loc[i, 'Order'])
            if order != -1:
                el = asspar.loc[i, 'DiffEl']
                int1 = asspar.loc[i, 'IntEl1']
                int2 = asspar.loc[i, 'IntEl2']
                k = "MQ_" + phase + "_" + el + "_in_" + int1 + "_" + int2 + "_" + str(order)
                d[k] = {}
                d[k]['Value_lowT'] = asspar.loc[i, 'MQlowT']
                d[k]['Value_highT'] = asspar.loc[i, 'MQhighT']
                d[k]['BP'] = asspar.loc[i, 'BP']
        return d

    @classmethod
    def tern_to_dict(cls, asspar, phase):
        """Save ternary results in dictionaries."""
        asspar.reset_index(inplace=True, drop=True)
        d = {}
        for i in range(len(asspar)):
            el = asspar.loc[i, 'DiffEl']
            int1 = asspar.loc[i, 'IntEl1']
            int2 = asspar.loc[i, 'IntEl2']
            int3 = asspar.loc[i, 'IntEl3']
            main_elem = asspar.loc[i, 'MainEl']
            k = "MQ_" + phase + "_" + el + "_in_" + int1 + "_" + int2 + "_" + int3 + "_main_" + main_elem + "_" + str(2)
            d[k] = {}
            d[k]['Value_lowT'] = asspar.loc[i, 'MQlowT']
            d[k]['Value_highT'] = asspar.loc[i, 'MQighT']
            d[k]['BP'] = asspar.loc[i, 'BP']
        return d


class InitRes:
    _PATH = os.getcwd()

    @classmethod
    def init_output(cls, phases: dict):
        """Initialize output directory."""
        new_path = InitRes._PATH + "/Results"
        General.create_dir(new_path)
        InitRes._set_path(new_path)

    @classmethod
    def _set_path(cls, path: str):
        InitRes._PATH = path

    @classmethod
    def get_path(cls):
        return InitRes._PATH



