#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import general as gf
from output import parameter as pp


class BinaryPlot:

    def __init__(self, diff_el, inter1, inter2, phase, data, ass_par_un, ass_par_bin, unary=False):
        self._el = diff_el
        self._inter1 = inter1
        self._inter2 = inter2
        self._al = self._get_alloy()
        self._phase = phase
        self._asspar_un = ass_par_un
        self._asspar_bin = ass_par_bin
        self._data = data
        self._path = gf.General.PATH
        self._el_id = self._al.index(self._el)
        self._unary = unary

    def main(self):
        self._init_output()
        self._init_vs_comp_plots()
        self._init_vs_temp_plots()

    def _get_alloy(self):
        """Creates an alloy list for the given alloy."""
        al = [self._el, self._inter1, self._inter2]
        al = list(set(al))
        al.sort()
        return al

    def _init_output(self):
        self._path = self._path + "/Results"
        gf.General.create_dir(self._path)
        if self._unary:
            self._path = self._path + "/Endmember_Results"
            gf.General.create_dir(self._path)
            self._path = self._path + "/" + self._phase
            gf.General.create_dir(self._path)
            self._path = self._path + "/" + self._el + "_in_" + self._inter2
            gf.General.create_dir(self._path)
        else:
            self._path = self._path + "/Binary_Interactions"
            gf.General.create_dir(self._path)
            self._path = self._path + "/" + self._phase
            gf.General.create_dir(self._path)
            self._path = self._path + "/" + self._el + "_in_" + self._inter1 + "_" + self._inter2
            gf.General.create_dir(self._path)

    def _init_vs_temp_plots(self):
        """Initializes plots against the temperature."""
        self._y_vs_temp_plot("invT", "lnD", refs_sort=False, shifted=True)
        self._y_vs_temp_plot("invT", "lnD", refs_sort=True, shifted=True)
        self._y_vs_temp_plot("invT", "lnD", refs_sort=True, shifted=False)

    def _init_vs_comp_plots(self):
        """Initializes plots against the composition."""
        self._y_vs_comp_plot("lnD", refs_sort=True, shifted=True)
        self._y_vs_comp_plot("lnD", refs_sort=False, shifted=True)
        self._y_vs_comp_plot("lnD", refs_sort=True, shifted=False)

    def _y_vs_comp_plot(self, y_var, refs_sort=False, shifted=True):
        """Plots the given y parameter against the composition."""
        if self._data is not None:
            refs = pd.DataFrame()
            if refs_sort:
                refs = list(gf.DataManipulator.get_list_of_references(self._data).Ref)
            temp_unique = list(set(self._data['Temp']))
            shi = 0
            shift = 0
            if shifted:
                shift = self._get_shift(y_var)
            for i, loc_temp in enumerate(temp_unique):
                if shifted:
                    shi = shift * i
                lab = str(loc_temp) + " [K]"
                if shifted:
                    lab = lab + "; shift: " + str("{:.1f}".format(shi))
                local_dat = self._data.loc[self._data['Temp']==loc_temp]
                local_dat.reset_index(drop=True, inplace=True)
                if refs_sort:
                    for ind,ref in enumerate(refs):
                        if ind == 0:
                            plt.plot(local_dat.loc[local_dat['Original'] == ref, self._el].astype(float),
                                 local_dat.loc[local_dat['Original'] == ref, y_var].astype(float) + shi, marker=pp.PlotParameter().get_marker(i), markersize=pp.PlotParameter.markersize*2,
                                linestyle='None', color=pp.PlotParameter().get_colour(i), label=lab)
                        else:
                            plt.plot(local_dat.loc[local_dat['Original'] == ref, self._el].astype(float),
                                 local_dat.loc[local_dat['Original'] == ref, y_var].astype(float) + shi, marker=pp.PlotParameter().get_marker(i), markersize=pp.PlotParameter.markersize*2,
                                linestyle='None', color=pp.PlotParameter().get_colour(i), label='_nolegend_')
                else:
                    plt.plot(local_dat.loc[:, self._el].astype(float), local_dat.loc[:, y_var].astype(float) + shi, marker=pp.PlotParameter().get_marker(i), markersize=pp.PlotParameter.markersize*2, linestyle='None',
                             color=pp.PlotParameter().get_colour(i), label=lab)
                if self._asspar_un is not None:
                    lab = '_nolegend_'
                    temp_data_list = gf.PandasRelated.create_bin_help_data_list(self._al, loc_temp, loc_temp)
                    temp_ass_par = self._asspar_un
                    if self._asspar_bin is not None:
                        temp_ass_par = dict(self._asspar_un, **self._asspar_bin)
                    res = gf.LnDfromAss(self._el, temp_data_list, self._phase, temp_ass_par).main()
                    res = pd.Series(res)
                    comp = temp_data_list[self._el]
                    plt.plot(comp, res + shi, color=pp.PlotParameter.get_colour(i), linewidth=1.0, label=lab)

            plt.xlabel('x(' + self._el + ')')
            plt.ylabel('ln(D[m²/s])')
            if shifted:
                plt.ylabel('ln(D[m²/s]) + shift')
            leg = plt.legend(loc='best', prop={'size': 6}, bbox_to_anchor=(1.0, 1.0))
            leg.get_frame().set_alpha(0.5)
            plt.title(self._el + ' in ' + "".join(self._al))
            tit = self._path + '/' + self._el + '_in_' + "".join(self._al) + "_" + y_var + "_vs_comp_at_different_T"
            if refs_sort:
                tit = tit + "_sorted_by_ref"
            if shifted:
                tit = tit + "_shifted"
            plt.savefig(tit + pp.PlotParameter.fpdf, dpi=pp.PlotParameter.res, bbox_inches='tight')
            plt.savefig(tit + pp.PlotParameter.fpng, dpi=pp.PlotParameter.res, bbox_inches='tight')
            plt.close()
        else:
            print("Only assessment results without data is not implemented yet.")

    def _y_vs_temp_plot(self, x_var, y_var, refs_sort=False, shifted= True):
        """Plots the given y parameter against the temperature."""
        # TODO: automatic check whethre all data were used
        # TODO: checken ob der shift zu groß ist und dadurch Trends unerkennbar werden
        #TODO: shift muss noch angepasst werden
        factor = pp.PlotParameter.invT_fact
        if self._data is not None:
            refs = pd.DataFrame()
            if refs_sort:
                refs = list(gf.DataManipulator.get_list_of_references(self._data).Ref)
            comp_unique = list(set(self._data[self._el].astype(float).tolist()))
            shi = 0
            shift = 0
            if shifted:
                shift = self._get_shift(y_var)
            for i, loc_comp in enumerate(comp_unique):
                if shifted:
                    shi = shift * i
                lab = str("{:.2f}".format(loc_comp*100)) + " [at.%]"
                if shifted:
                    lab = lab + "; shift: " + str("{:.1f}".format(shi))
                local_dat = self._data.loc[self._data[self._el]==loc_comp]
                local_dat.reset_index(drop=True, inplace=True)
                if refs_sort:
                    for ind,ref in enumerate(refs):
                        if ind==0:
                            plt.plot(local_dat.loc[local_dat['Original'] == ref, x_var].astype(float)*factor,
                                 local_dat.loc[local_dat['Original'] == ref, y_var].astype(float) + shi, marker=pp.PlotParameter().get_marker(i), markersize=pp.PlotParameter.markersize*2,
                                linestyle='None', color=pp.PlotParameter().get_colour(i), label=lab)
                        else:
                            plt.plot(local_dat.loc[local_dat['Original'] == ref, x_var].astype(float)*factor,
                                 local_dat.loc[local_dat['Original'] == ref, y_var].astype(float) + shi, marker=pp.PlotParameter().get_marker(i), markersize=pp.PlotParameter.markersize*2,
                                linestyle='None', color=pp.PlotParameter().get_colour(i), label='_nolegend_')
                else:
                    plt.plot(local_dat.loc[:, x_var].astype(float)*factor,
                                 local_dat.loc[:, y_var].astype(float) + shi, marker=pp.PlotParameter().get_marker(i), markersize=pp.PlotParameter.markersize*2, linestyle='None',
                             color=pp.PlotParameter().get_colour(i), label=lab)

                if self._asspar_un is not None:
                    lab = '_nolegend_'
                    temp_data_list = gf.PandasRelated.create_bin_help_data_list(self._al, self._data['Temp'].min(), self._data['Temp'].max(), cmin=loc_comp, cmax=loc_comp)
                    temp_ass_par = self._asspar_un
                    if self._asspar_bin is not None:
                        temp_ass_par = dict(self._asspar_un, **self._asspar_bin)
                    res = gf.LnDfromAss(self._el, temp_data_list, self._phase, temp_ass_par).main()
                    res = pd.Series(res)
                    plt.plot(1.0/temp_data_list['Temp']*factor, res + shi, linewidth=1.0, color=pp.PlotParameter.get_colour(i), label=lab)

            xlab = str(factor) + '/T [1/K]'
            plt.xlabel(xlab)
            plt.ylabel('ln(D[m²/s])')
            if shifted:
                plt.ylabel('ln(D[m²/s]) + shift')

            leg = plt.legend(loc='best', prop={'size': 6}, bbox_to_anchor=(1, 1.0))
            leg.get_frame().set_alpha(0.5)
            plt.title(self._el + ' in ' + "".join(self._al))
            tit = self._path + '/' + self._el + '_in_' + "".join(self._al) + "_" + y_var + "_vs_" + x_var
            if refs_sort:
                tit = tit + "_sorted_by_ref"
            if shifted:
                tit = tit + "_shifted"
            plt.savefig(tit + pp.PlotParameter.fpdf, dpi=pp.PlotParameter.res, bbox_inches='tight')
            plt.savefig(tit + pp.PlotParameter.fpng, dpi=pp.PlotParameter.res, bbox_inches='tight')
            plt.close()

    def _get_shift(self, y_var):
        """Evaluates which shift is needed, depending on the element."""
        if y_var == "D":
            return pp.PlotParameter.shift_D
        else:
            return pp.PlotParameter.shift_lnD
