#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import general as gf

from db import ass_res as ar
from setup import Setup as setup


class DDBWriter:
    """Handles writing DDB files for given systems."""

    def __init__(self):
        self._file = self._init_file()
        self._alloy = setup.ELEMENTS
        self._phase_list = setup.PHASES
        self._ph = None
        self._T_min = 273.0
        self._T_max = 6000.0
        self._asspar_un = ar.GetAssRes().get_un()
        self._asspar_bin = ar.GetAssRes().get_bin()
        self._asspar_tern = ar.GetAssRes().get_tern()

    def main(self):
        """Handles the createion of a ddb file."""
        self._add_comment()
        self._define_elements()
        self._define_phase()
        self._create_DDB()
        self._file.close()

    def _init_file(self):
        """Initializes the output file for the general information.
        Args: None
        Returns: None
        """
        path = gf.General.PATH + "/Results"
        filename = path + "/PyMob_mobilities.ddb"
        file = open(filename, "w")
        return file

    def _define_phase(self):
        """Writes the sublattice occupation for the given phases to the file"""
        header = "$ ------------------------------------------------------------------- \n$ Phase " \
                 "definitions \n$ \n"
        self._file.write(header)
        for ph in self._phase_list:
            if ph == "FCC_A1" or ph == "FCC":
                d = "PHASE FCC_A1 %A 2 1 1 ! \n"
                self._file.write(d)
                elem = "CONST FCC_A1 : "
                for j in range(len(self._alloy)):
                    elem = elem + self._alloy[j] + " "
                elem = elem + ": VA% : !\n"
                self._file.write(elem)
                self._file.write("\n")
            else:
                print(ph + " does not exist yet. Add it in the DDB writer file.")
        end = "$ -------------------------------------------------------------------- \n  \n \n \n"
        self._file.write(end)

    def _add_comment(self):
        """Adds the initial comment for the database file"""
        text = "$ Copyright holder and editor:  Katrin Abrahams (katrin.abrahams@rub.de)\n"
        self._file.write(text)
        self._file.write("\n")

    def _define_elements(self):
        """Writes specific parameters for the elements into the file"""
        header = "$Element     Standard state   mass [g/mol]     H298-H0        S298    \n"
        self._file.write(header)
        # the values given here are taken from Dinsdale SGTE data, the entries are Standard State and mass
        el_def_DINS = {}
        el_def_DINS['VA'] = "VACUUM", 0.0, 0.0, 0.0
        el_def_DINS['CO'] = "HCP_A3", 58.933, 4765.567, 30.0400
        el_def_DINS['CR'] = "BCC_A2", 51.996, 4050.0, 23.5429
        el_def_DINS['FE'] = "BCC_A2", 55.847, 4489.0, 27.2797
        el_def_DINS['MN'] = "BCC_A12", 54.9380, 4995.696, 32.2206
        el_def_DINS['NI'] = "FCC_A1", 58.69, 4787.0, 29.7955

        el = 'VA'
        col = "ELEMENT " + el + "        " + el_def_DINS[el][0] + "        " + str(el_def_DINS[el][1]) + "        " + \
              str(el_def_DINS[el][2]) + "        " + str(el_def_DINS[el][3]) + " ! \n"
        self._file.write(col)

        for i in range(len(self._alloy)):
            el = self._alloy[i]
            try:
                el_def_DINS[el]
            except:
                print(el + " has to be added to the element definitions in the DDB writer")
                exit(1)
            else:
                col = "ELEMENT " + el + "        " + el_def_DINS[el][0] + "        " + str(el_def_DINS[el][1]) + "        " + \
                      str(el_def_DINS[el][2]) + "        " + str(el_def_DINS[el][3]) + " ! \n"
                self._file.write(col)
        self._file.write("\n")
        self._file.write("\n")
        m = " DEFAULT_COMMAND DEF_SYS_ELEMENT VA ! \n"
        self._file.write(m)
        m = " TYPE-DEF % SEQ * ! \n"
        self._file.write(m)

    def _create_DDB(self):
        """Handles DDB creation"""
        for ph in self._phase_list:
            self._ph = ph
            header = "$ ------------------------------------------------------------------- $ \n$  FCC " \
                     " \n$ ------------------------------------------------------------------- $ \n"
            self._file.write(header)
            if not self._asspar_un.empty:
                self._add_default()
                self._add_unary_functions()
            if not self._asspar_bin.empty:
                self._add_binary_parameters()
            if not self._asspar_tern.empty:
                self._add_ternary_parameters()

    def _add_ternary_parameters(self):
        header = " \n $ Ternary interaction values \n"
        self._file.write(header)
        for i in range(len(self._asspar_tern)):
            if self._asspar_tern.loc[i, 'MQlowT'] is not None:
                el_list = [self._asspar_tern.loc[i, 'IntEl1'], self._asspar_tern.loc[i, 'IntEl2'], self._asspar_tern.loc[i, 'IntEl3']]
                el_list.sort()
                main_el = self._asspar_tern.loc[i, 'MainEl']
                if main_el in el_list:
                    index = el_list.index(main_el)
                    MQ_lowT = round(self._asspar_tern.loc[i, 'MQlowT'], 4)
                    MQ_highT = round(self._asspar_tern.loc[i, 'MQhighT'], 4)
                    diff_el = self._asspar_tern.loc[i, 'DiffusingElement']
                    if MQ_highT != MQ_lowT and MQ_highT != 0 and MQ_lowT != 0:
                        bp = round(self._asspar_tern.loc[i, 'BP'],4)
                        func = "PARAMETER MQ(" + self._ph + "&" + diff_el + "," + el_list[0] + "," + el_list[1] + "," + el_list[2]  + ":*;" + str(index) + ") " \
                       + str(self._T_min) + " " + str("{:f}".format(MQ_lowT)) + "; " + str(bp) + " Y \n"
                        self._file.write(func)
                        func = "PARAMETER MQ(" + self._ph + "&" + diff_el + "," + el_list[0] + "," + el_list[1] + "," + el_list[2]  + ":*;" + str(index) + ") " \
                       + str(bp) + " " + str("{:f}".format(MQ_highT)) + "; " + str(self._T_max) + " N !\n"
                        self._file.write(func)
                    elif MQ_lowT != 0:
                        func = "PARAMETER MQ(" + self._ph + "&" + diff_el + "," + el_list[0] + "," + el_list[1] + "," + el_list[2]  + ":*;" + str(index) + ") " \
                       + str(self._T_min) + " " + str("{:f}".format(MQ_lowT)) + "; " + str(self._T_max) + " N !\n"
                        self._file.write(func)

    def _add_binary_parameters(self):
        header = " \n $ Binary interaction values \n"
        self._file.write(header)
        for i in range(len(self._asspar_bin)):
            if self._asspar_bin.loc[i, 'MQlowT'] is not None:
                MQ_lowT = round(self._asspar_bin.loc[i, 'MQlowT'], 4)
                MQ_highT = round(self._asspar_bin.loc[i, 'MQhighT'], 4)
                el = self._asspar_bin.loc[i, 'DiffusingElement']
                inter1 = self._asspar_bin.loc[i, 'IntEl1']
                inter2 = self._asspar_bin.loc[i, 'IntEl2']
                order = self._asspar_bin.loc[i, 'Order']
                if MQ_highT != MQ_lowT and order != -1:
                    bp = round(self._asspar_bin.loc[i, 'BP'],4)
                    func = "PARAMETER MQ(" + self._ph + "&" + el + "," + inter1 + "," + inter2 + ":*;" + str(order) + ") " \
                       + str(self._T_min) + " " + str("{:f}".format(MQ_lowT)) + "; " + str(bp) + " Y \n"
                    self._file.write(func)
                    func = "PARAMETER MQ(" + self._ph + "&" + el + "," + inter1 + "," + inter2 + ":*;" + str(order) + ") " \
                       + str(bp) + " " + str("{:f}".format(MQ_highT)) + "; " + str(self._T_max) + " N !\n"
                    self._file.write(func)
                elif order != -1:
                    func = "PARAMETER MQ(" + self._ph + "&" + el + "," + inter1 + "," + inter2 + ":*;" + str(order) + ") " \
                       + str(self._T_min) + " " + str("{:f}".format(MQ_lowT)) + "; " + str(self._T_max) + " N !\n"
                    self._file.write(func)

    def _add_default(self):
        """Adds the default functions for every element"""
        header = " \n $ Default Values \n"
        self._file.write(header)
        for i in range(len(self._asspar_un)):
            mat = self._asspar_un.loc[i, 'Matrix']
            el = self._asspar_un.loc[i, 'DiffusingElement']
            if mat == el:
                MQ_lowT = round(self._asspar_un.loc[i, 'MQlowT']*(-1),4)
                MF_lowT = round(self._asspar_un.loc[i, 'MFlowT'],6)
                MQ_highT = round(self._asspar_un.loc[i, 'MQlowT']*(-1),4)
                MF_highT = round(self._asspar_un.loc[i, 'MFlowT'],6)
                if MQ_lowT != MQ_highT:
                    bp = round(self._asspar_un.loc[i, 'BP'],4)
                    par = "PARAMETER MQ(" + self._ph + "&" + el + ",*) " + str(self._T_min) + "  " + str("{:f}".format(MQ_lowT)) + ";  " + str(bp) + " Y \n"
                    self._file.write(par)
                    par = "PARAMETER MF(" + self._ph + "&" + el + ",*) " + str(self._T_min) + "  " + "R*T*" + str(MF_lowT) + ";  " + str(bp) + " N !\n"
                    self._file.write(par)
                    par = "PARAMETER MQ(" + self._ph + "&" + el + ",*) " + str(bp) + "  " + str("{:f}".format(MQ_highT)) + ";  " + str(self._T_max) + " Y \n"
                    self._file.write(par)
                    par = "PARAMETER MF(" + self._ph + "&" + el + ",*) " + str(bp) + "  " + "R*T*" + str(MF_highT) + ";  " + str(self._T_max) + " N !\n"
                    self._file.write(par)
                else:
                    par = "PARAMETER MQ(" + self._ph + "&" + el + ",*) " + str(self._T_min) + "  " + str("{:f}".format(MQ_lowT)) + ";  " + str(self._T_max) + " N !\n"
                    self._file.write(par)
                    par = "PARAMETER MF(" + self._ph + "&" + el + ",*) " + str(self._T_min) + "  " + "R*T*" + str(MF_lowT) + ";  " + str(self._T_max) + " N !\n"
                    self._file.write(par)

    def _add_unary_functions(self):
        """Adds the assessed unary data"""
        header = " \n $ Emd-member Values \n"
        self._file.write(header)
        for i in range(len(self._asspar_un)):
            mat = self._asspar_un.loc[i, 'Matrix']
            el = self._asspar_un.loc[i, 'DiffusingElement']
            MQ_lowT = round(self._asspar_un.loc[i, 'MQlowT'] * (-1),4)
            MF_lowT = round(self._asspar_un.loc[i, 'MFlowT'],6)
            MQ_highT = round(self._asspar_un.loc[i, 'MQlowT'] * (-1),4)
            MF_highT = round(self._asspar_un.loc[i, 'MFlowT'],6)
            if MQ_lowT != MQ_highT:
                bp  = round(self._asspar_un.loc[i, 'BP'],4)
                par = "PARAMETER MQ(" + self._ph + "&" + el + "," +  mat + ":*) " + str(self._T_min) + "  " + str(
                    "{:f}".format(MQ_lowT)) + ";  " + str(bp) + " N !\n"
                self._file.write(par)
                par = "PARAMETER MF(" + self._ph + "&" + el + "," +  mat + ":*) " + str(self._T_min) + "  " + "R*T*" + str(
                    MF_lowT) + ";  " + str(bp) + " N !\n"
                self._file.write(par)
                par = "PARAMETER MQ(" + self._ph + "&" + el + "," +  mat + ":*) " + str(bp) + "  " + str(
                    "{:f}".format(MQ_highT)) + ";  " + str(self._T_max) + " N !\n"
                self._file.write(par)
                par = "PARAMETER MF(" + self._ph + "&" + el + "," +  mat + ":*) " + str(bp) + "  " + "R*T*" + str(
                    MF_highT) + ";  " + str(self._T_max) + " N !\n"
                self._file.write(par)
            else:
                par = "PARAMETER MQ(" + self._ph + "&" + el + "," +  mat + ":*) " + str(self._T_min) + "  " + str(
                    "{:f}".format(MQ_lowT)) + ";  " + str(self._T_max) + " N !\n"
                self._file.write(par)
                par = "PARAMETER MF(" + self._ph + "&" + el + "," +  mat + ":*) " + str(self._T_min) + "  " + "R*T*" + str(
                    MF_lowT) + ";  " + str(self._T_max) + " N !\n"
                self._file.write(par)
