#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import pandas as pd

import general as gf
from db import raw_data as rd


class AnalyseRD:
    """Analyses the raw data stored in the SQL databases, once giving a general overview and then it creates for every
    phase a file with more details."""

    __slots__ = ('_file', '_data', '_temp_data', '_list_phases', '_diff_element_list', '_ph', '_el', '_phase_file')

    def __init__(self):
        self._file = self._init_file()
        self._data = self._get_data()
        self._temp_data = pd.DataFrame()
        self._list_phases = []
        self._diff_element_list = []
        self._ph = None
        self._el = None
        self._phase_file = None

    def analyse(self):
        """Main function handeling the analysis of the raw data
        Args: None
        Returns: None
        """
        self._list_phases = list(set(self._data.Phase.tolist()))
        if self._list_phases:
            self._list_phases.sort()
            self._print_phase_overview()
            for ph in self._list_phases:
                self._ph = ph
                self._temp_data = self._data.loc[self._data['Phase'] == self._ph].copy()
                self._diff_element_list = list(set(self._temp_data['DiffusingElement'].tolist()))
                self._create_phase_header(self._file)
                self._init_phase_file()
                self._fill_phase_file()
                self._phase_file.close()
        self._file.close()

    def _init_file(self):
        """Initializes the output file for the general information.
        Args: None
        Returns: None
        """
        path = gf.General.PATH + "/Results"
        filename = path + "/RawDataStatistics.txt"
        file = open(filename, "w")
        header = "Available Raw Data in PyMob \n \n"
        file.write(header)
        return file

    def _init_phase_file(self):
        """Initializes the output file for the more detailed phase files.
        Args: None
        Returns: None
        """
        path = gf.General.PATH + "/Results"
        filename = path + "/" + self._ph + ".txt"
        self._phase_file = open(filename, "w")
        self._create_phase_header(self._phase_file)

    def _print_phase_overview(self):
        """Initializes the output file for every phase.
        Args: None
        Returns: None
        """
        header = "Distribution of data for different phases: \n"
        self._file.write(header)
        for ph in self._list_phases:
            number_data = len(self._data.loc[self._data['Phase']==ph])
            text = ph.title() + ": " + str(number_data) + "\n"
            self._file.write(text)
        self._add_break(self._file)

    def _create_phase_header(self, file):
        """Creates the header in the files for the phases.
        Args: None
        Returns: None
        """
        header = "Distribution of data in " + self._ph + ": \n"
        file.write(header)
        number_data = len(self._data.loc[self._data['Phase'] == self._ph])
        text = "Total number of available data: " + str(number_data) + "\n"
        file.write(text)
        subheader = "Data for the following diffusing elements: \n"
        file.write(subheader)
        for el in self._diff_element_list:
            number_data = len(self._data.loc[self._data['DiffusingElement'] == el])
            text = el.title() + ": " + str(number_data) + "\n"
            file.write(text)
        self._add_break(file)

    def _get_data(self) -> pd.DataFrame:
        """Collects all raw data from the database.
        Args: None
        Returns: None
        """
        data = rd.GetData().main()
        data = gf.Conv().arr_par_into_d(data)
        data = gf.DataManipulator().prepare_data(data)
        return data

    def _add_break(self, file):
        """Adds a break. Used to distinguish passages from each other in the created files.
        Args: None
        Returns: None
        """
        text = "==================================================== \n \n"
        file.write(text)

    def _fill_phase_file(self):
        """Adds the detailed information in the files for every phase.
        Args: None
        Returns: None
        """
        for el in self._diff_element_list:
            self._el = el
            header = "Data for " + self._el.title() + " diffusion: \n"
            self._phase_file.write(header)
            data_by_el = self._temp_data.loc[self._data['DiffusingElement'] == self._el].copy()
            data_by_el.reset_index(inplace=True, drop=True)
            al_list = gf.PandasRelated.get_alloy_list(data_by_el)
            for ind, al in enumerate(al_list):
                dat = data_by_el.loc[data_by_el[al].isnull().sum(axis=1) == 0]
                refs = gf.DataManipulator.get_list_of_references(dat)
                for i in range(len(al)):
                    al[i] = al[i].title()
                al = ''.join(al)
                text = "Diffusion data in " + al + ": " + str(len(dat)) + "; "
                self._phase_file.write(text)
                subtext = "References: "
                self._phase_file.write(subtext)
                for i in range(len(refs)):
                    subsubtext = refs.loc[i, 'Ref'] + " (" + str(refs.loc[i, 'DP']) + ") "
                    self._phase_file.write(subsubtext)
                final_text = "\n"
                self._phase_file.write(final_text)
            self._add_break(self._phase_file)






