#
#   This file is part of the PyMob software project.
#   For more details visit www.icams.de
#
#   Created: 2020
#
#   Copyright (c) 2020-2021 Interdisciplinary Centre for Advanced Materials
#                 Simulation (ICAMS). Ruhr-Universitaet Bochum. Germany
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import numpy as np
import typing


class ElementProperties:

    #_properties = ['atomic number','mass', 'mp', 'curie_temp', 'standard state', 'H298-H0', 'S298']
    _symbols = [
        'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca',
        'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y',
        'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce',
        'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir',
        'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm',
        'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn', 'Nh', 'Fl',
        'Mc', 'Lv', 'Ts', 'Og']
    _symbols = [s.upper() for s in _symbols]
    _names = [
        'Hydrogen', 'Helium', 'Lithium', 'Beryllium', 'Boron', 'Carbon', 'Nitrogen', 'Oxygen', 'Fluorine', 'Neon',
        'Sodium', 'Magnesium', 'Aluminium', 'Silicon', 'Phosphorus', 'Sulfur', 'Chlorine', 'Argon', 'Potassium',
        'Calcium', 'Scandium', 'Titanium', 'Vanadium', 'Chromium', 'Manganese', 'Iron', 'Cobalt', 'Nickel', 'Copper',
        'Zinc', 'Gallium', 'Germanium', 'Arsenic', 'Selenium', 'Bromine', 'Krypton', 'Rubidium', 'Strontium', 'Yttrium',
        'Zirconium', 'Niobium', 'Molybdenum', 'Technetium', 'Ruthenium', 'Rhodium', 'Palladium', 'Silver', 'Cadmium',
        'Indium', 'Tin', 'Antimony', 'Tellurium', 'Iodine', 'Xenon', 'Caesium', 'Barium', 'Lanthanum', 'Cerium',
        'Praseodymium', 'Neodymium', 'Promethium', 'Samarium', 'Europium', 'Gadolinium', 'Terbium', 'Dysprosium',
        'Holmium', 'Erbium', 'Thulium', 'Ytterbium', 'Lutetium', 'Hafnium', 'Tantalum', 'Tungsten', 'Rhenium', 'Osmium',
        'Iridium', 'Platinum', 'Gold', 'Mercury', 'Thallium', 'Lead', 'Bismuth', 'Polonium', 'Astatine', 'Radon',
        'Francium', 'Radium', 'Actinium', 'Thorium', 'Protactinium', 'Uranium', 'Neptunium', 'Plutonium', 'Americium',
        'Curium', 'Berkelium', 'Californium', 'Einsteinium', 'Fermium', 'Mendelevium', 'Nobelium', 'Lawrencium',
        'Rutherfordium', 'Dubnium', 'Seaborgium', 'Bohrium', 'Hassium', 'Meitnerium', 'Darmstadtium', 'Roentgenium',
        'Corpernicium', 'Nihonium', 'Flerovium', 'Moscovium', 'Livermorium', 'Tennessine', 'Oganesson']
    _atomic_numbers = range(1,119)
    _mass = [
        1.008, 4.002602, 6.94, 9.0121831, 10.81, 12.011, 14.007, 15.999, 18.998403163, 20.1797, 22.98976928, 24.305,
        26.9815384, 28.085, 30.973761998, 32.06, 35.45, 39.95, 39.0983, 40.078, 44.955908, 47.867, 50.9415, 51.9961,
        54.938043, 55.845, 58.933194, 58.6934, 63.546, 65.38, 69.723, 72.630, 74.921595, 78.971, 79.904, 83.798, 85.4678,
        87.62, 88.90584, 91.224, 92.90637, 95.95, 98, 101.07,102.90549,106.42, 107.8682, 112.414, 114.818, 118.710,
        121.760, 127.60, 126.90447,131.293, 132.90545196, 137.327, 138.90547, 140.116, 140.90766, 144.242, 145, 150.36,
        151.964, 157.25, 158.925354, 162.500, 164.930328, 167.259, 168.934218, 173.045, 174.9668, 178.49, 180.94788,
        183.84, 186.207, 190.23, 192.217, 195.084, 196.966570, 200.592, 204.38, 207.2, 208.98040, 209, 210, 222, 223,
        226, 227, 232.0377, 231.03588, 238.02891, 237, 244, 243, 247, 247, 251, 252, 257, 258, 259, 266, 267, 268, 269,
        270, 270, 278, 281, 282, 285, 286, 289, 290, 293, 294, 294
    ]
    # im K
    _melting_point = [14.01, np.nan, 453.69, 1560, 2349, 3800, 63.15, 54.36, 53.53, 24.56, 370.87, 923, 933.47, 1687,
                      317.30, 388.36, 171.6, 83.80, 336.53, 1115, 1814, 1941, 2183, 2180, 1519, 1811, 1768, 1728,
                      1357.77, 692.88, 302.9146, 1211.40, 1090, 453, 265.8, 115.79, 312.46, 1050, 1799, 2128, 2750,
                      2896, 2430, 2607, 2237, 1828.05, 1234.93, 594.22, 429.75, 505.08, 903.78, 722.66, 386.85, 161.4,
                      301.59, 1000, 1193, 1068, 1208, 1297, 1315, 1345, 1099, 1585, 1629, 1680, 1734, 1802, 1818, 1097,
                      1925, 2506, 3290, 3695, 3459, 3306, 2719, 2041.4, 1337.33, 234.43, 577, 600.61, 544.7, 527, 575,
                      202, 281, 973, 1323, 2115, 1841, 1405.3, 917, 912.5, 1449, 1613, 1259, 1173, 1133, 1125, 1100,
                      1100, 1900, 2400, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, 283, 700, np.nan, 700,
                      700, 700, 320
                      ]

    _properties = [{'name': name, 'atomic_number': atomic_number, 'mass': mass, 'melting_point': melting_point} for
                   name, atomic_number, mass, melting_point in zip(_names, _atomic_numbers, _mass, _melting_point)]
    _elements = dict(zip(_symbols, _properties))

    @classmethod
    def get_long_name(cls, el: str) -> typing.Optional[str]:
        el = el.upper().strip()
        if el in ElementProperties._elements:
            return ElementProperties._elements[el]['name']
        else:
            return None

    @classmethod
    def get_mass(cls, el: str) -> typing.Optional[float]:
        el = el.upper().strip()
        if el in ElementProperties._elements:
            return ElementProperties._elements[el]['mass']
        else:
            return None

    @classmethod
    def get_melting_point(cls, el: str) -> typing.Optional[float, np.nan]:
        el = el.upper().strip()
        if el in ElementProperties._elements:
            return ElementProperties._elements[el]['melting_point']
        else:
            return None


class Phases:
    """Holds and modifies the dict of phases. The key is the phase name and the value references the phase class."""
    _PHASE_SYNONYMS = {"FCC": "FCC_A1", "FCC_A1": "FCC"}

    __slots__ = '_phase_dict'

    def __init__(self):
        self._phase_dict = []

    def get_phases(self):
        return self._phase_dict


if __name__ == '__main__':
    print(ElementProperties._elements)
    print(ElementProperties.get_long_name('ni '))
    print(ElementProperties.get_mass('ni '))
    print(ElementProperties.get_melting_point('ni '))




